#include  "global.h"
#include  <cstdlib>
#include  "lexer.h" 
#include  "axis.h"
#include <sstream>
#include <iomanip>
#include  "matreduc.h"
#include  "bitmap.h"
#include  "lattice.h"
#include  "bitvector.h"
#include  "bitvector.h"
#include  "permutations.h"

namespace atlas { namespace interpreter {

Id_table* global_id_table=nullptr; 

overload_table* global_overload_table=nullptr;

std::ostream* output_stream= &std::cout;

bool clean=true;

int verbosity=0;

const shared_bool global_false = std::make_shared<bool_value>(false);
const shared_bool global_true  = std::make_shared<bool_value>(true);

namespace {


void add_overload(id_type id, shared_function&& f, type_expr&& type)
{
  size_t old_n=global_overload_table->variants(id).size();
std::ostringstream type_string;
  type_string << type;
    // save type |type| as string before moving from it
  global_overload_table->add(id,std::move(f),std::move(type));
    // insert or replace table entry
  size_t n=global_overload_table->variants(id).size();
  if (n==old_n)
    *output_stream << "Redefined ";
  else if (n==1)
    *output_stream << "Defined ";
  else
    *output_stream << "Added definition [" << n << "] of ";
  *output_stream << main_hash_table->name_of(id) << ": "
            << type_string.str();
  *output_stream << std::endl;
}

void handle
  (const std::exception& err,const id_pat& pat, int phase, int overload)
{ static const char* message[3] = {"not executed","interrupted","failed"};
  std::cerr << err.what() << "\n  Command 'set " << pat << "' "
            << message[phase];
  if (phase<2)
    std::cerr << ", nothing " << (overload<=1 ? "defin" : "overload") << "ed";
  std::cerr << ".\n";
clean=false;
  reset_evaluator(); main_input_buffer->close_includes();
}
void do_global_set(id_pat&& pat, const expr& rhs, int overload)
{ size_t n_id=count_identifiers(pat);
  int phase; // needs to be declared outside the |try|, is used in |catch|
  try
  { phase=0; // type check
    expression_ptr e;
    type_expr t=analyse_types(rhs,e);
    if (not pattern_type(pat).specialise(t))
      
      { std::ostringstream o;
        o << "Type " << t 
          << " of right hand side does not match required pattern "
          << pattern_type(pat);
        throw program_error(o.str());
      }
    
    { if (overload==2 and t.kind!=function_type)
        throw program_error("Cannot set operator to a non function value");
    }
    layer b(n_id);
    thread_bindings(pat,t,b,false); // match identifiers and their future types


    phase=1; // evaluation of right hand side
  e->eval();

    phase=2; // actual definition of identifiers
    std::vector<shared_value> v;
    v.reserve(n_id);
    thread_components(pat,pop_value(),std::back_inserter(v));
     // associate values with identifiers
    auto v_it = v.cbegin();
    for (auto it = b.begin(); it!=b.end(); ++it,++v_it)
    { assert(v_it!=v.cend());
      
      { unsigned int input_level = main_input_buffer->include_depth();
        *output_stream << std::setw(2*input_level) << "";
      }
      if (overload==0 or it->second.kind!=function_type)
      
      { *output_stream << (b.is_const(it) ? "Constant " : "Variable ") 
                       << main_hash_table->name_of(it->first)
                       << ": " << it->second;
        if (global_id_table->present(it->first))
        { bool is_const;
          *output_stream << " (overriding previous instance, which had type "
                    << *global_id_table->type_of(it->first,is_const);
          if (is_const)
            *output_stream << " (constant)";
          *output_stream << ')';
        }
        *output_stream << std::endl;
        global_id_table->add
          (it->first,std::move(*v_it),std::move(it->second),b.is_const(it));
      }
      else
      
      { shared_function f = std::dynamic_pointer_cast<const function_base>(*v_it);
        if (f.get()==nullptr)
          throw logic_error("Non-function value found with function type");
        add_overload(it->first,std::move(f),std::move(it->second));
      }
    }
  }
  
  catch (const program_error& err)
  { assert(phase!=1); handle(err,pat,phase,overload); }
  catch (const runtime_error& err)
  { assert(phase==1); handle(err,pat,phase,overload); }
  catch (const logic_error& err)
  { std::cerr << "Unexpected error: ";
    handle(err,pat,phase,overload);
  }
  catch (const std::exception& err)
  { handle(err,pat,phase,overload); }
}

void rational_convert() // convert integer to rational (with denominator~1)
{ int i = get<int_value>()->val;
    push_value(std::make_shared<rat_value>(Rational(i)));
}

void ratlist_ratvec_convert() // convert list of rationals to rational vector
{ shared_row r = get<row_value>();
  int_Vector numer(r->val.size()),denom(r->val.size());
  unsigned int d=1;
  for (size_t i=0; i<r->val.size(); ++i)
  // collect numerators and denominators separately
  { Rational frac = force<rat_value>(r->val[i].get())->val;
    numer[i]=frac.numerator();
    denom[i]=frac.denominator();
    d=arithmetic::lcm(d,denom[i]); // and compute the least common denominator
  }
  for (size_t i=0; i<r->val.size(); ++i)
    numer[i]*= d/denom[i]; // adjust numerators to common denominator

  push_value(std::make_shared<rational_vector_value>(std::move(numer),d));
  // normalises
}

void ratvec_ratlist_convert() // convert rational vector to list of rationals
{ shared_rational_vector rv = get<rational_vector_value>();
  own_row result = std::make_shared<row_value>(rv->val.size());
  for (size_t i=0; i<rv->val.size(); ++i)
  { Rational q(rv->val.numerator()[i],rv->val.denominator());
    result->val[i] = std::make_shared<rat_value>(q.normalize());
  }
  push_value(std::move(result)); // here |std::move| avoids ref-count updates
}

void vec_ratvec_convert() // convert vector to rational vector
{ own_vector v = get_own<vector_value>();
  push_value(std::make_shared<rational_vector_value>
    (RatWeight(std::move(v->val),1)));
}

int_Vector row_to_weight(const row_value& r)
{ int_Vector result(r.val.size());
  for(size_t i=0; i<r.val.size(); ++i)
    result[i]=force<int_value>(r.val[i].get())->val;
  return result;
}

void intlist_vector_convert()
{ shared_row r = get<row_value>();
  push_value(std::make_shared<vector_value>(row_to_weight(*r)));
}

void intlist_ratvec_convert()
{ shared_row r = get<row_value>();
  push_value(std::make_shared<rational_vector_value>
    (RatWeight(row_to_weight(*r),1)));
}

void veclist_matrix_convert()
{ shared_row r = get<row_value>();
  if (r->val.size()==0)
    throw runtime_error("Cannot convert empty list of vectors to matrix");

  size_t n = force<vector_value>(r->val[0].get())->val.size();
  own_matrix m = std::make_shared<matrix_value>(int_Matrix(n,r->val.size()));
  for(size_t j=0; j<r->val.size(); ++j)
  { const int_Vector& col = force<vector_value>(r->val[j].get())->val;
    if (col.size()!=n)
      throw runtime_error("Vector sizes differ in conversion to matrix");

    m->val.set_column(j,col);
  }
  push_value(std::move(m));
}

void intlistlist_matrix_convert()
{ shared_row r = get<row_value>();
  if (r->val.size()==0)
    throw runtime_error("Cannot convert empty list of lists to matrix");

  size_t n = force<row_value>(r->val[0].get())->val.size();
  own_matrix m = std::make_shared<matrix_value>(int_Matrix(n,r->val.size()));
  for(size_t j=0; j<r->val.size(); ++j)
  { int_Vector col = row_to_weight(*force<row_value>(r->val[j].get()));
    if (col.size()!=n)
      throw runtime_error("List sizes differ in conversion to matrix");

    m->val.set_column(j,col);
  }
  push_value(std::move(m));
}

own_row weight_to_row(const int_Vector& v)
{ own_row result = std::make_shared<row_value>(v.size());
  for(size_t i=0; i<v.size(); ++i)
    result->val[i]=std::make_shared<int_value>(v[i]);
  return result;
}

void vector_intlist_convert()
{ shared_vector v = get<vector_value>();
  push_value(weight_to_row(v->val));
}

void matrix_veclist_convert()
{ shared_matrix m=get<matrix_value>();
  own_row result = std::make_shared<row_value>(m->val.numColumns());
  for(size_t i=0; i<m->val.numColumns(); ++i)
    result->val[i]=std::make_shared<vector_value>(m->val.column(i));
  push_value(std::move(result));
}

void matrix_intlistlist_convert()
{ shared_matrix m=get<matrix_value>();
  own_row result = std::make_shared<row_value>(m->val.numColumns());
  for(size_t i=0; i<m->val.numColumns(); ++i)
    result->val[i]= weight_to_row(m->val.column(i));

  push_value(std::move(result));
}

void plus_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(i+j));
}

void minus_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(i-j));
}

void times_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(i*j));
}

void divide_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (j==0) throw runtime_error("Division by zero");
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>
     (j>0 ? arithmetic::divide(i,j) : -arithmetic::divide(i,-j)));
}

void modulo_wrapper(expression_base::level l)
{ int  j=get<int_value>()->val; int i=get<int_value>()->val;
  if (j==0) throw runtime_error("Modulo zero");
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>
      (arithmetic::remainder(i,std::abs(j))));
}

void divmod_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (j==0) throw runtime_error("DivMod by zero");
  if (l!=expression_base::no_value)
  { push_value(std::make_shared<int_value>
     (j>0 ? arithmetic::divide(i,j) : -arithmetic::divide(i,-j)));
    push_value(std::make_shared<int_value>
      (arithmetic::remainder(i,std::abs(j))));
    if (l==expression_base::single_value)
      wrap_tuple<2>();
  }
}

void unary_minus_wrapper(expression_base::level l)
{ int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(-i));
}

void power_wrapper(expression_base::level l)
{ static shared_int one = std::make_shared<int_value>(1);
  static shared_int minus_one  = std::make_shared<int_value>(-1);
int n=get<int_value>()->val; int i=get<int_value>()->val;
  if (std::abs(i)!=1 and n<0)
    throw runtime_error("Negative power of integer");
  if (l==expression_base::no_value)
    return;

  if (i==1)
  { push_value(one);
      return;
  }
  if (i==-1)
  { push_value(n%2==0 ? one : minus_one);
      return;
  }

  push_value(std::make_shared<int_value>(arithmetic::power(i,n)));
}

void fraction_wrapper(expression_base::level l)
{ int d=get<int_value>()->val; int n=get<int_value>()->val;
  if (d==0) throw runtime_error("fraction with zero denominator");
  if (d<0) { d=-d; n=-n; } // ensure denominator is positive
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(Rational(n,d)));
}


void unfraction_wrapper(expression_base::level l)
{ Rational q=get<rat_value>()->val;
  if (l!=expression_base::no_value)
  { push_value(std::make_shared<int_value>(q.numerator()));
    push_value(std::make_shared<int_value>(q.denominator()));
    if (l==expression_base::single_value)
      wrap_tuple<2>();
  }
}

void rat_plus_int_wrapper(expression_base::level l)
{ int i=get<int_value>()->val;
  own_rat q=get_own<rat_value>();
  if (l==expression_base::no_value)
    return;
  q->val+=i;
  push_value(q);
}
void rat_minus_int_wrapper(expression_base::level l)
{ int i=get<int_value>()->val;
  own_rat q=get_own<rat_value>();
  if (l==expression_base::no_value)
    return;
  q->val-=i;
  push_value(q);
}
void rat_times_int_wrapper(expression_base::level l)
{ int i=get<int_value>()->val;
  own_rat q=get_own<rat_value>();
  if (l==expression_base::no_value)
    return;
  q->val*=i;
  push_value(q);
}
void rat_divide_int_wrapper(expression_base::level l)
{ int i=get<int_value>()->val;
  own_rat q=get_own<rat_value>();
  if (i==0)
    throw runtime_error("Rational division by zero");
  if (l==expression_base::no_value)
    return;
  q->val/=i;
  push_value(q);
}
void rat_modulo_int_wrapper(expression_base::level l)
{ int i=get<int_value>()->val;
  own_rat q=get_own<rat_value>();
  if (i==0)
    throw runtime_error("Rational modulo zero");
  if (l==expression_base::no_value)
    return;
  q->val%=i;
  push_value(q);
}

void rat_plus_wrapper(expression_base::level l)
{ Rational j=get<rat_value>()->val;
  Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(i+j));
}
void rat_minus_wrapper(expression_base::level l)
{ Rational j=get<rat_value>()->val;
  Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(i-j));
}

void rat_times_wrapper(expression_base::level l)
{ Rational j=get<rat_value>()->val;
  Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(i*j));
}

void rat_divide_wrapper(expression_base::level l)
{ Rational j=get<rat_value>()->val;
  Rational i=get<rat_value>()->val;
  if (j.numerator()==0)
    throw runtime_error("Rational division by zero");
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(i/j));
}
void rat_modulo_wrapper(expression_base::level l)
{ Rational j=get<rat_value>()->val;
  Rational i=get<rat_value>()->val;
  if (j.numerator()==0)
    throw runtime_error("Rational modulo zero");
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(i%j));
}

void rat_unary_minus_wrapper(expression_base::level l)
{ Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(Rational(0)-i)); }

void rat_inverse_wrapper(expression_base::level l)
{ Rational i=get<rat_value>()->val;
  if (i.numerator()==0)
    throw runtime_error("Inverse of zero");
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(Rational(1)/i)); }

void rat_power_wrapper(expression_base::level l)
{ int n=get<int_value>()->val; Rational b=get<rat_value>()->val;
  if (b.numerator()==0 and n<0)
    throw runtime_error("Negative power of zero");
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rat_value>(b.power(n)));
}

void int_unary_eq_wrapper(expression_base::level l)
{ int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(i==0));
}
void int_unary_neq_wrapper(expression_base::level l)
{ int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(i!=0));
}
void int_non_negative_wrapper(expression_base::level l)
{ int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(i>=0));
}
void int_positive_wrapper(expression_base::level l)
{ int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(i>0));
}
void int_non_positive_wrapper(expression_base::level l)
{ int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(i<=0));
}
void int_negative_wrapper(expression_base::level l)
{ int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(i<0));
}

void int_eq_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(i==j));
}

void int_neq_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(i!=j));
}

void int_less_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(i<j));
}

void int_lesseq_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(i<=j));
}

void int_greater_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(i>j));
}

void int_greatereq_wrapper(expression_base::level l)
{ int j=get<int_value>()->val; int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(i>=j));
}

void rat_unary_eq_wrapper(expression_base::level l)
{ Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(i.numerator()==0));
}
void rat_unary_neq_wrapper(expression_base::level l)
{ Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(i.numerator()!=0));
}
void rat_non_negative_wrapper(expression_base::level l)
{ Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(i.numerator()>=0));
}
void rat_positive_wrapper(expression_base::level l)
{ Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(i.numerator()>0));
}
void rat_non_positive_wrapper(expression_base::level l)
{ Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(i.numerator()<=0));
}
void rat_negative_wrapper(expression_base::level l)
{ Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(i.numerator()<0));
}

void rat_eq_wrapper(expression_base::level l)
{ Rational j=get<rat_value>()->val; Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(i==j));
}

void rat_neq_wrapper(expression_base::level l)
{ Rational j=get<rat_value>()->val; Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(i!=j));
}

void rat_less_wrapper(expression_base::level l)
{ Rational j=get<rat_value>()->val; Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(i<j));
}

void rat_lesseq_wrapper(expression_base::level l)
{ Rational j=get<rat_value>()->val; Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(i<=j));
}

void rat_greater_wrapper(expression_base::level l)
{ Rational j=get<rat_value>()->val; Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(i>j));
}

void rat_greatereq_wrapper(expression_base::level l)
{ Rational j=get<rat_value>()->val; Rational i=get<rat_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(i>=j));
}

void equiv_wrapper(expression_base::level l)
{ bool a=get<bool_value>()->val; bool b=get<bool_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(a==b));
}

void inequiv_wrapper(expression_base::level l)
{ bool a=get<bool_value>()->val; bool b=get<bool_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(a!=b));
}

void string_unary_eq_wrapper(expression_base::level l)
{ shared_string i=get<string_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val.empty()));
}
void string_unary_neq_wrapper(expression_base::level l)
{ shared_string i=get<string_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val.size()>0));
}
void string_eq_wrapper(expression_base::level l)
{ shared_string j=get<string_value>(); shared_string i=get<string_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val==j->val));
}
void string_neq_wrapper(expression_base::level l)
{ shared_string j=get<string_value>(); shared_string i=get<string_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val!=j->val));
}

void string_less_wrapper(expression_base::level l)
{ shared_string j=get<string_value>(); shared_string i=get<string_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val<j->val));
}
void string_leq_wrapper(expression_base::level l)
{ shared_string j=get<string_value>(); shared_string i=get<string_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val<=j->val));
}
void string_greater_wrapper(expression_base::level l)
{ shared_string j=get<string_value>(); shared_string i=get<string_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val>j->val));
}
void string_geq_wrapper(expression_base::level l)
{ shared_string j=get<string_value>(); shared_string i=get<string_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val>=j->val));
}

void int_format_wrapper(expression_base::level l)
{ int n=get<int_value>()->val;
  std::ostringstream o; o<<n;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<string_value>(o.str()));
}

void string_concatenate_wrapper(expression_base::level l)
{ shared_string b=get<string_value>(); shared_string a=get<string_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<string_value>(a->val+b->val));
}

void concatenate_strings_wrapper(expression_base::level l)
{ shared_row arg=get<row_value>();
  if (l==expression_base::no_value)
    return;
  const std::vector<shared_value>& x=arg->val;
  std::vector<const std::string*> p; p.reserve(x.size());
  for (auto it=x.cbegin(); it!=x.cend(); ++it)
    p.push_back(&force<string_value>(it->get())->val);
  size_t s=0;
  for (auto it=p.cbegin(); it!=p.cend(); ++it)
    s+=(*it)->size();
  std::string result(s,char()); auto dst=result.begin();
  for (auto it=p.cbegin(); it!=p.cend(); ++it)
    dst=std::copy((*it)->begin(),(*it)->end(),dst);
  assert(dst==result.end());
  push_value(std::make_shared<string_value>(std::move(result)));
}

void string_to_ascii_wrapper(expression_base::level l)
{ shared_string c=get<string_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>
      (c->val.size()==0 ? -1 : (unsigned char)c->val[0]));
}

void ascii_char_wrapper(expression_base::level l)
{ int c=get<int_value>()->val;
  if ((c<' ' and c!='\n') or c>'~')
    throw runtime_error("Value "+str(c)+" out of range");
  if (l!=expression_base::no_value)
    push_value(std::make_shared<string_value>(std::string(1,c)));
}

void matrix_bounds_wrapper(expression_base::level l)
{ shared_matrix m=get<matrix_value>();
  if (l==expression_base::no_value)
    return;
  push_value(std::make_shared<int_value>(m->val.numRows()));
  push_value(std::make_shared<int_value>(m->val.numColumns()));
  if (l==expression_base::single_value)
    wrap_tuple<2>();
}

void vector_suffix_wrapper(expression_base::level l)
{ int e=get<int_value>()->val;
  own_vector r=get_own<vector_value>();
  if (l!=expression_base::no_value)
  { r->val.push_back(e);
    push_value(r);
  }
}

void vector_prefix_wrapper(expression_base::level l)
{ own_vector r=get_own<vector_value>();
  int e=get<int_value>()->val;
  if (l!=expression_base::no_value)
  { r->val.insert(r->val.begin(),e);
    push_value(r);
  }
}

void join_vectors_wrapper(expression_base::level l)
{ shared_vector y=get<vector_value>();
  shared_vector x=get<vector_value>();
  if (l!=expression_base::no_value)
  { own_vector result = std::make_shared<vector_value>(int_Vector());
    result->val.reserve(x->val.size()+y->val.size());
    result->val.insert(result->val.end(),x->val.begin(),x->val.end());
    result->val.insert(result->val.end(),y->val.begin(),y->val.end());
    push_value(std::move(result));
  }

}

void join_vector_row_wrapper(expression_base::level l)
{ shared_row arg=get<row_value>();
  if (l==expression_base::no_value)
    return;
  const std::vector<shared_value>& x=arg->val;
  std::vector<const int_Vector*> p; p.reserve(x.size());
  for (auto it=x.cbegin(); it!=x.cend(); ++it)
    p.push_back(&force<vector_value>(it->get())->val);
  size_t s=0;
  for (auto it=p.cbegin(); it!=p.cend(); ++it)
    s+=(*it)->size();
  int_Vector result(s);
  auto dst=result.begin();
  for (auto it=p.cbegin(); it!=p.cend(); ++it)
    dst=std::copy((*it)->cbegin(),(*it)->cend(),dst);
  assert(dst==result.end());
  push_value(std::make_shared<vector_value>(std::move(result)));
}

void vec_unary_eq_wrapper(expression_base::level l)
{ shared_vector i=get<vector_value>();
  if (l==expression_base::no_value)
    return;
  const auto end=i->val.end();
  for (auto it=i->val.begin(); it!=end; ++it)
    if (*it!=0)
    { push_value(whether(false));
      return; }
  push_value(whether(true));
}
void vec_unary_neq_wrapper(expression_base::level l)
{ shared_vector i=get<vector_value>();
  if (l==expression_base::no_value)
    return;
  const auto end=i->val.end();
  for (auto it=i->val.begin(); it!=end; ++it)
    if (*it!=0)
    { push_value(whether(true));
      return; }
  push_value(whether(false));
}
void vec_eq_wrapper(expression_base::level l)
{ shared_vector j=get<vector_value>(); shared_vector i=get<vector_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val==j->val));
}
void vec_neq_wrapper(expression_base::level l)
{ shared_vector j=get<vector_value>(); shared_vector i=get<vector_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val!=j->val));
}

void vec_non_negative_wrapper(expression_base::level l)
{ shared_vector v = get<vector_value>();
  if (l==expression_base::no_value)
    return;
  bool OK=true;
  for (auto it=v->val.begin(); it!=v->val.end(); ++it)
    if (*it<0)
    { OK=false; break; }
  push_value(whether(OK));
}
void vec_positive_wrapper(expression_base::level l)
{ shared_vector v = get<vector_value>();
  if (l==expression_base::no_value)
    return;
  bool OK=true;
  for (auto it=v->val.begin(); it!=v->val.end(); ++it)
    if (*it<=0)
    { OK=false; break; }
  push_value(whether(OK));
}

void ratvec_unary_eq_wrapper(expression_base::level l)
{ shared_rational_vector v=get<rational_vector_value>();
  if (l==expression_base::no_value)
    return;
  push_value(whether(v->val.isZero()));
}
void ratvec_unary_neq_wrapper(expression_base::level l)
{ shared_rational_vector v=get<rational_vector_value>();
  if (l==expression_base::no_value)
    return;
  push_value(whether(not v->val.isZero()));
}
void ratvec_eq_wrapper(expression_base::level l)
{ shared_rational_vector w=get<rational_vector_value>();
  shared_rational_vector v=get<rational_vector_value>();
  if (l!=expression_base::no_value)
    push_value(whether(v->val==w->val));
}
void ratvec_neq_wrapper(expression_base::level l)
{ shared_rational_vector w=get<rational_vector_value>();
  shared_rational_vector v=get<rational_vector_value>();
  if (l!=expression_base::no_value)
    push_value(whether(v->val!=w->val));
}

void ratvec_non_negative_wrapper(expression_base::level l)
{ shared_rational_vector v = get<rational_vector_value>();
  if (l==expression_base::no_value)
    return;
  bool OK=true;
  for (auto it=v->val.numerator().begin(); it!=v->val.numerator().end(); ++it)
    if (*it<0)
    { OK=false; break; }
  push_value(whether(OK));
}
void ratvec_positive_wrapper(expression_base::level l)
{ shared_rational_vector v = get<rational_vector_value>();
  if (l==expression_base::no_value)
    return;
  bool OK=true;
  for (auto it=v->val.numerator().begin(); it!=v->val.numerator().end(); ++it)
    if (*it<=0)
    { OK=false; break; }
  push_value(whether(OK));
}

void mat_unary_eq_wrapper(expression_base::level l)
{ shared_matrix i=get<matrix_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val.is_zero()));
}
void mat_unary_neq_wrapper(expression_base::level l)
{ shared_matrix i=get<matrix_value>();
  if (l!=expression_base::no_value)
    push_value(whether(not i->val.is_zero()));
}
void mat_eq_wrapper(expression_base::level l)
{ shared_matrix j=get<matrix_value>(); shared_matrix i=get<matrix_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val==j->val));
}
void mat_neq_wrapper(expression_base::level l)
{ shared_matrix j=get<matrix_value>(); shared_matrix i=get<matrix_value>();
  if (l!=expression_base::no_value)
    push_value(whether(i->val!=j->val));
}

void check_size (size_t a, size_t b)
{ if (a==b)
    return;
  std::ostringstream s; s<< "Size mismatch " << a << ":" << b;
  throw runtime_error(s.str());
}

void vec_plus_wrapper(expression_base::level l)
{ own_vector v1= get_own<vector_value>();
  shared_vector v0= get<vector_value>();
  check_size(v0->val.size(),v1->val.size());
  if (l==expression_base::no_value)
    return;
  v1->val += v0->val;
  push_value(std::move(v1));
}
void vec_minus_wrapper(expression_base::level l)
{ own_vector v1= get_own<vector_value>();
  shared_vector v0= get<vector_value>();
  check_size(v0->val.size(),v1->val.size());
  if (l==expression_base::no_value)
    return;
  v1->val.negate_add(v0->val);
  push_value(std::move(v1));
}

void vec_unary_minus_wrapper(expression_base::level l)
{ own_vector v = get_own<vector_value>();
  if (l==expression_base::no_value)
    return;
  v->val.negate();
  push_value(std::move(v));
}

void vec_times_int_wrapper(expression_base::level l)
{ int i= get<int_value>()->val;
  own_vector v= get_own<vector_value>();
  if (l==expression_base::no_value)
    return;
  v->val *= i;
  push_value(std::move(v));
}
void vec_divide_int_wrapper(expression_base::level l)
{ int i= get<int_value>()->val;
  own_vector v= get_own<vector_value>();
  if (i==0)
    throw runtime_error("Vector division by 0");
  if (l==expression_base::no_value)
    return;
  divide(v->val,i);
  push_value(std::move(v));
}
void vec_modulo_int_wrapper(expression_base::level l)
{ int i= get<int_value>()->val;
  own_vector v= get_own<vector_value>();
  if (i==0)
    throw runtime_error("Vector modulo 0");
  if (l==expression_base::no_value)
    return;
  v->val %= i;
  push_value(std::move(v));
}

void vv_prod_wrapper(expression_base::level l)
{ shared_vector w=get<vector_value>();
  shared_vector v=get<vector_value>();
  check_size (v->val.size(),w->val.size());
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(v->val.dot(w->val)));
}

void flex_add_wrapper(expression_base::level l)
{ shared_vector v1= get<vector_value>();
  shared_vector v0= get<vector_value>();
  if (l==expression_base::no_value)
    return;
  const int_Vector& V0=v0->val;
  size_t i0=V0.size();
const int_Vector& V1=v1->val;
  size_t i1=V1.size();
  while (i0>0 and V0[i0-1]==0)
    --i0;
  while (i1>0 and V1[i1-1]==0)
    --i1;
  if (i0==i1) // equal size case
  { while (i0>0 and V0[i0-1]+V1[i0-1]==0)
      --i0; // skip leading zeros in result
    int_Vector result(i0);
    while (i0-->0)
      result[i0]=V0[i0]+V1[i0];
    push_value(std::make_shared<vector_value>(std::move(result)));
    return;
  }
  shared_vector& larger = *(i0>i1 ? &v0 : &v1); // unequal size case
  if (larger.unique()) // then we can grab the larger vector
  { own_vector result =
      std::const_pointer_cast<vector_value>(std::move(larger));
    if (i0>i1)
    {
      result->val.resize(i0);
      while (i1-->0)
        result->val[i1]+=V1[i1]; // add |V1| to result
    }
    else
    {
      result->val.resize(i1);
      while (i0-->0)
        result->val[i0]+=V0[i0]; // add |V0| to result
    }
    push_value(std::move(result));
  }
  else // we must allocate
  {
    int_Vector result(std::max(i0,i1));
    while (i0>i1)
      --i0,result[i0]=V0[i0]; // copy excess of |V0|
    while (i1>i0)
      --i1,result[i1]=V1[i1]; // copy excess of |V1|
    while (i0-->0)
      result[i0]=V0[i0]+V1[i0];
    push_value(std::make_shared<vector_value>(std::move(result)));
  }
}

void flex_sub_wrapper(expression_base::level l)
{ shared_vector v1= get<vector_value>();
  shared_vector v0= get<vector_value>();
  if (l==expression_base::no_value)
    return;
  const int_Vector& V0=v0->val;
  size_t i0=V0.size();
const int_Vector& V1=v1->val;
  size_t i1=V1.size();
  while (i0>0 and V0[i0-1]==0)
    --i0;
  while (i1>0 and V1[i1-1]==0)
    --i1;
  if (i0==i1) // equal size case
  { while (i0>0 and V0[i0-1]==V1[i0-1])
      --i0; // skip leading zeros in result
    int_Vector result(i0);
    while (i0-->0)
      result[i0]=V0[i0]-V1[i0];
    push_value(std::make_shared<vector_value>(std::move(result)));
    return;
  }
  shared_vector& larger = *(i0>i1 ? &v0 : &v1); // unequal size case
  if (larger.unique()) // then we can grab the larger vector
  { own_vector result =
      std::const_pointer_cast<vector_value>(std::move(larger));
    if (i0>i1)
    {
      result->val.resize(i0);
      while (i1-->0)
        result->val[i1]-=V1[i1]; // subtract |V1| from result
    }
    else
    {
      result->val.resize(i1);
      while (i1-->i0)
        result->val[i1]=-result->val[i1]; // negate top part of result
      while (i0-->0)
        result->val[i0]=V0[i0]-result->val[i0]; // negate result, add |V0|
    }
    push_value(std::move(result));
  }
  else // we must allocate
  {
    int_Vector result(std::max(i0,i1));
    while (i0>i1)
      --i0,result[i0]=V0[i0]; // copy excess of |V0|
    while (i1>i0)
      --i1,result[i1]=-V1[i1]; // copy excess of |V1|, negated
    while (i0-->0)
      result[i0]=V0[i0]-V1[i0];
    push_value(std::make_shared<vector_value>(std::move(result)));
  }
}

void vector_convolve_wrapper(expression_base::level l)
{ shared_vector v1= get<vector_value>();
  shared_vector v0= get<vector_value>();
  if (l==expression_base::no_value)
    return;
  const int_Vector& V0=v0->val;
  size_t i0=V0.size();
const int_Vector& V1=v1->val;
  size_t i1=V1.size();
  while (i0>0 and V0[i0-1]==0)
    --i0;
  while (i1>0 and V1[i1-1]==0)
    --i1;
  if (i0==0 or i1==0)
{ push_value(std::make_shared<vector_value>(int_Vector(0)));
    return;
  }
  own_vector result = std::make_shared<vector_value>(int_Vector(i0+i1-1));
  int_Vector& r = result->val;
  size_t i=i0,j=0; int V1j=V1[0], V0l=V0[i0-1];
  while (i-->0)
    r[i] = V0[i]*V1j; // copy |V0|, multiplied by lowest (constant) term of |V1|
  while (++j<i1)
  { r[(i=i0-1)+j] = V0l*(V1j=V1[j]); // copy top term of |V0| times next of |V1|
    while (i-->0)
      r[i+j] += V0[i]*V1j; // add remainder of multiple of |V0|, shifted |j|
  }
  push_value(std::move(result));
}

void vector_div_wrapper(expression_base::level l)
{ int n=get<int_value>()->val;
  own_vector v=get_own<vector_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rational_vector_value>
      (std::move(v->val),n)); // throws if |n==0|
}

void ratvec_unfraction_wrapper(expression_base::level l)
{ shared_rational_vector v = get<rational_vector_value>();
  if (l!=expression_base::no_value)
  { Weight num(v->val.numerator().begin(),v->val.numerator().end()); // convert
    push_value(std::make_shared<vector_value>(std::move(num)));
    push_value(std::make_shared<int_value>(v->val.denominator()));
    if (l==expression_base::single_value)
      wrap_tuple<2>();
  }
}

void ratvec_plus_wrapper(expression_base::level l)
{ shared_rational_vector v1= get<rational_vector_value>();
  shared_rational_vector v0= get<rational_vector_value>();
  check_size(v0->val.size(),v1->val.size());
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rational_vector_value>(v0->val+v1->val));
}
void ratvec_minus_wrapper(expression_base::level l)
{ shared_rational_vector v1= get<rational_vector_value>();
  shared_rational_vector v0= get<rational_vector_value>();
  check_size(v0->val.size(),v1->val.size());
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rational_vector_value>(v0->val-v1->val));
}
void ratvec_unary_minus_wrapper(expression_base::level l)
{ shared_rational_vector v = get<rational_vector_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rational_vector_value>(-v->val));
}

void ratvec_times_int_wrapper(expression_base::level l)
{ int i= get<int_value>()->val;
  own_rational_vector v= get_own<rational_vector_value>();
  if (l==expression_base::no_value)
    return;
  (v->val *= i).normalize();
  push_value(v);
}
void ratvec_divide_int_wrapper(expression_base::level l)
{ int i= get<int_value>()->val;
  own_rational_vector v= get_own<rational_vector_value>();
  if (i==0)
    throw runtime_error("Rational vector division by 0");
  if (l==expression_base::no_value)
    return;
  (v->val /= i).normalize();
  push_value(v);
}
void ratvec_modulo_int_wrapper(expression_base::level l)
{ int i= get<int_value>()->val;
  own_rational_vector v= get_own<rational_vector_value>();
  if (i==0)
    throw runtime_error("Rational vector modulo 0");
  if (l==expression_base::no_value)
    return;
  v->val %= i;
  push_value(v);
}


void ratvec_times_rat_wrapper(expression_base::level l)
{ Rational r= get<rat_value>()->val;
  own_rational_vector v= get_own<rational_vector_value>();
  if (l==expression_base::no_value)
    return;
  (v->val *= r).normalize();
  push_value(v);
}
void ratvec_divide_rat_wrapper(expression_base::level l)
{ Rational r= get<rat_value>()->val;
  own_rational_vector v= get_own<rational_vector_value>();
  if (r.numerator()==0)
    throw runtime_error("Rational vector division by 0");
  if (l==expression_base::no_value)
    return;
  (v->val /= r).normalize();
  push_value(v);
}

void mat_plus_int_wrapper(expression_base::level l)
{ int i= get<int_value>()->val;
  own_matrix m= get_own<matrix_value>();
  if (l==expression_base::no_value)
    return;
  m->val += i;
  push_value(m);
}
void mat_minus_int_wrapper(expression_base::level l)
{ int i= get<int_value>()->val;
  own_matrix m= get_own<matrix_value>();
  if (l==expression_base::no_value)
    return;
  m->val += -i;
  push_value(m);
}

void int_plus_mat_wrapper(expression_base::level l)
{ own_matrix m= get_own<matrix_value>();
  int i= get<int_value>()->val;
  if (l==expression_base::no_value)
    return;
  m->val += i;
  push_value(m);
}
void int_minus_mat_wrapper(expression_base::level l)
{ own_matrix m= get_own<matrix_value>();
  int i= get<int_value>()->val;
  if (l==expression_base::no_value)
    return;
  m->val.negate();
  m->val += i;
  push_value(m);
}

void mat_plus_mat_wrapper(expression_base::level l)
{ own_matrix b= get_own<matrix_value>();
  shared_matrix a= get<matrix_value>();
  check_size(a->val.numRows(),b->val.numRows());
  check_size(a->val.numColumns(),b->val.numColumns());
  if (l==expression_base::no_value)
    return;
  b->val += a->val;
  push_value(b);
}

void mat_minus_mat_wrapper(expression_base::level l)
{
  shared_matrix b= get<matrix_value>();
  own_matrix a= get_own<matrix_value>();
  check_size(a->val.numRows(),b->val.numRows());
  check_size(a->val.numColumns(),b->val.numColumns());
  if (l==expression_base::no_value)
    return;
  a->val -= b->val;
  push_value(a);
}

void mv_prod_wrapper(expression_base::level l)
{ shared_vector v=get<vector_value>();
  shared_matrix m=get<matrix_value>();
  if (m->val.numColumns()!=v->val.size())
    throw runtime_error(std::string("Size mismatch ")
     + str(m->val.numColumns()) + ":" + str(v->val.size()));
  if (l!=expression_base::no_value)
    push_value(std::make_shared<vector_value>(m->val*v->val));
}

void mrv_prod_wrapper(expression_base::level l)
{ shared_rational_vector v=get<rational_vector_value>();
  shared_matrix m=get<matrix_value>();
  if (m->val.numColumns()!=v->val.size())
    throw runtime_error(std::string("Size mismatch ")
     + str(m->val.numColumns()) + ":" + str(v->val.size()));
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rational_vector_value>(m->val*v->val));
}

void vm_prod_wrapper(expression_base::level l)
{ shared_matrix m=get<matrix_value>(); // right factor
  shared_vector v=get<vector_value>(); // left factor
  if (v->val.size()!=m->val.numRows())
  { std::ostringstream s;
    s<< "Size mismatch " << v->val.size() << ":" << m->val.numRows();
    throw runtime_error(s.str());
  }
  if (l!=expression_base::no_value)
    push_value(std::make_shared<vector_value>(m->val.right_prod(v->val)));
}

void rvm_prod_wrapper(expression_base::level l)
{ shared_matrix m=get<matrix_value>();
  shared_rational_vector v=get<rational_vector_value>();
  if (v->val.size()!=m->val.numRows())
    throw runtime_error(std::string("Size mismatch ")
     + str(v->val.size()) + ":" + str(m->val.numRows()));
  if (l!=expression_base::no_value)
    push_value(std::make_shared<rational_vector_value>(v->val*m->val));
}

void null_vec_wrapper(expression_base::level l)
{ int n=get<int_value>()->val;
  if (n<0)
    throw runtime_error("Negative size for vector: "+str(n));
  if (l!=expression_base::no_value)
    push_value(std::make_shared<vector_value>(int_Vector(n,0)));
}
 void null_mat_wrapper(expression_base::level l)
{ int n=get<int_value>()->val;
  int m=get<int_value>()->val;
  if (m<0)
    throw runtime_error("Negative number of rows: "+str(m));
  if (n<0)
    throw runtime_error("Negative number of columns: "+str(n));
  if (l!=expression_base::no_value)
    push_value(std::make_shared<matrix_value> (int_Matrix(m,n,0)));
}
void transpose_vec_wrapper(expression_base::level l)
{ shared_vector v=get<vector_value>();
  if (l!=expression_base::no_value)
  { own_matrix m = std::make_shared<matrix_value>(int_Matrix(1,v->val.size()));
    for (size_t j=0; j<v->val.size(); ++j)
      m->val(0,j)=v->val[j];
    push_value(std::move(m));
  }
}

void diagonal_wrapper(expression_base::level l)
{ shared_vector d=get<vector_value>();
  if (l==expression_base::no_value)
    return;
  size_t n=d->val.size();
  own_matrix m = std::make_shared<matrix_value>(int_Matrix(n,n,0));
  for (size_t i=0; i<n; ++i)
    m->val(i,i)=d->val[i];
  push_value(std::move(m));
}

void stack_rows_wrapper(expression_base::level l)
{ shared_row r = get<row_value>();
  size_t n = r->val.size();
  std::vector<const int_Vector*> row(n);
  size_t width=0; // maximal length of vectors
  for(size_t i=0; i<n; ++i)
  { row[i] = & force<vector_value>(r->val[i].get())->val;
    if (row[i]->size()>width)
      width=row[i]->size();
  }

  if (l==expression_base::no_value)
    return;

  own_matrix m = std::make_shared<matrix_value>(int_Matrix(n,width,0));
  for(size_t i=0; i<n; ++i)
    for (size_t j=0; j<row[i]->size(); ++j)
      m->val(i,j)=(*row[i])[j];
  push_value(std::move(m));
}

void combine_columns_wrapper(expression_base::level l)
{ shared_row r = get<row_value>();
  int n = get<int_value>()->val;
  if (n<0)
    throw runtime_error("Negative number "+str(n)+" of rows requested");

  own_matrix m = std::make_shared<matrix_value>(int_Matrix(n,r->val.size()));
  for(size_t j=0; j<r->val.size(); ++j)
  { const int_Vector& col = force<vector_value>(r->val[j].get())->val;
    if (col.size()!=size_t(n))
      throw runtime_error("Column "+str(j)+" size "+str(col.size())
          +" does not match specified size "+str(n));

    m->val.set_column(j,col);
  }
  if (l!=expression_base::no_value)
    push_value(std::move(m));
}

void combine_rows_wrapper(expression_base::level l)
{ shared_row r =get<row_value>();
  int n = get<int_value>()->val;
  if (n<0)
    throw runtime_error("Negative number "+str(n)+" of columns requested");

  own_matrix m = std::make_shared<matrix_value>(int_Matrix(r->val.size(),n));
  for(size_t i=0; i<r->val.size(); ++i)
  { const int_Vector& row = force<vector_value>(r->val[i].get())->val;
    if (row.size()!=size_t(n))
      throw runtime_error("Row "+str(i)+" size "+str(row.size())
          +" does not match specified size "+str(n));

    m->val.set_row(i,row);
  }
  if (l!=expression_base::no_value)
    push_value(std::move(m));
}

template <bool transpose, bool negate>
void transform_copy
  (unsigned flags,
   const int_Matrix& src, 
   int lwb_r, int upb_r,
   int lwb_c, int upb_c,
   int_Matrix& dst)
{ if ((flags&0x1)==0) // no reversal of rows
  { if ((flags&0x2)==0) // no reversal of rows or columns
      for (int i=0, k=lwb_r; k<upb_r; ++i, ++k)
	for (int j=0, l=lwb_c; l<upb_c; ++j, ++l)
	  *(transpose ? &dst(j,i) : &dst(i,j)) = (negate ? -1 : 1) * src(k,l);
    else // reversal of columns only
      for (int i=0, k=lwb_r; k<upb_r; ++i, ++k)
	for (int j=0, l=upb_c; l-->lwb_c; ++j)
	  *(transpose ? &dst(j,i) : &dst(i,j)) = (negate ? -1 : 1) * src(k,l);
  }
  else // reversal of rows
  { if ((flags&0x2)==0) // reversal of rows only
      for (int i=0, k=upb_r; k-->lwb_r; ++i)
	for (int j=0, l=lwb_c; l<upb_c; ++j, ++l)
	  *(transpose ? &dst(j,i) : &dst(i,j)) = (negate ? -1 : 1) * src(k,l);
    else // reversal of rows and columns
      for (int i=0, k=upb_r; k-->lwb_r; ++i)
	for (int j=0, l=upb_c; l-->lwb_c; ++j)
	  *(transpose ? &dst(j,i) : &dst(i,j)) = (negate ? -1 : 1) * src(k,l);
  }
}

void swiss_matrix_knife_wrapper(expression_base::level lev)
{ int l = get<int_value>()->val;
  int j = get<int_value>()->val;
  int k = get<int_value>()->val;
  int i = get<int_value>()->val;
  shared_matrix src = get<matrix_value>();
  const int_Matrix& A = src->val;
  BitSet<8> flags (get<int_value>()->val);

  int m = A.numRows(); int n= A.numColumns();
  int lwb_r = flags[1] ? m-i : i;
  int upb_r = flags[2] ? m-k : k;
  int lwb_c = flags[4] ? n-j : j;
  int upb_c = flags[5] ? n-l : l;
  if (lwb_r<0 or upb_r>m or lwb_c<0 or upb_c>n)
    
    { std::ostringstream o;
      o << "Range exceeds bounds: ";
      if (lwb_r<0 or upb_r>m)
        if (upb_r>m)
          if (lwb_r<0)
            o << "both row bounds " << lwb_r << ':' << upb_r;
          else
            o << "upper row bound " << upb_r;
        else
          o << "lower row bound " << lwb_r;
      if ((lwb_r<0 or upb_r>m) and (lwb_c<0 or upb_c>n))
        o << " and ";
      if (lwb_c<0 or upb_c>n)
        if (upb_c>n)
          if (lwb_c<0)
            o << "both column bounds " << lwb_c << ':' << upb_c;
          else
            o << "upper column bound " << upb_c;
        else
          o << "lower column bound " << lwb_c;
      o << " out of range, actual bounds 0:" << m << ", 0:" << n;
      throw runtime_error(o.str());
    }

  if (lev==expression_base::no_value)
    return;
  
  if (lwb_r>upb_r)
    upb_r = lwb_r;
  if (lwb_c>upb_c)
    upb_c = lwb_c;
  unsigned r_size = upb_r - lwb_r;
  unsigned c_size = upb_c - lwb_c;
  if (flags[6]) // transpose
    std::swap(r_size,c_size);
  own_matrix result(std::make_shared<matrix_value>(int_Matrix(r_size,c_size)));
  unsigned rev_flags = static_cast<unsigned>(flags[0])*0x1
                     ^ static_cast<unsigned>(flags[3])*0x2;
  
  if (flags[6])
  { if (flags[7])
      transform_copy<true,true>
        (rev_flags,A,lwb_r,upb_r,lwb_c,upb_c,result->val);
    else
      transform_copy<true,false>
        (rev_flags,A,lwb_r,upb_r,lwb_c,upb_c,result->val);
  }
  else
  { if (flags[7])
      transform_copy<false,true>
        (rev_flags,A,lwb_r,upb_r,lwb_c,upb_c,result->val);
    else
      transform_copy<false,false>
        (rev_flags,A,lwb_r,upb_r,lwb_c,upb_c,result->val);
  }
  push_value(std::move(result));
}

void echelon_wrapper(expression_base::level l)
{ own_matrix M=get_own<matrix_value>();
  if (l!=expression_base::no_value)
  { BitMap pivots=matreduc::column_echelon(M->val);
    push_value(M);
    own_row p_list = std::make_shared<row_value>(0);
    p_list->val.reserve(pivots.size());
    for (BitMap::iterator it=pivots.begin(); it(); ++it)
      p_list->val.push_back(std::make_shared<int_value>(*it));
    push_value(std::move(p_list));
    if (l==expression_base::single_value)
      wrap_tuple<2>();
  }
}

void diagonalize_wrapper(expression_base::level l)
{ shared_matrix M=get<matrix_value>();
  if (l!=expression_base::no_value)
  { own_matrix row = std::make_shared<matrix_value>(int_Matrix());
    own_matrix column = std::make_shared<matrix_value>(int_Matrix());
    own_vector diagonal = std::make_shared<vector_value>
       (matreduc::diagonalise(M->val,row->val,column->val));
    push_value(std::move(diagonal));
    push_value(std::move(row));
    push_value(std::move(column));
    if (l==expression_base::single_value)
      wrap_tuple<3>();
  }
}

void adapted_basis_wrapper(expression_base::level l)
{ shared_matrix M=get<matrix_value>();
  if (l!=expression_base::no_value)
  { own_vector diagonal = std::make_shared<vector_value>(std::vector<int>());
    push_value(std::make_shared<matrix_value>
      (matreduc::adapted_basis(M->val,diagonal->val)));
    push_value(std::move(diagonal));
    if (l==expression_base::single_value)
      wrap_tuple<2>();
  }
}

void kernel_wrapper(expression_base::level l)
{ shared_matrix M=get<matrix_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<matrix_value>(lattice::kernel(M->val)));
}

void eigen_lattice_wrapper(expression_base::level l)
{ int eigen_value = get<int_value>()->val;
  shared_matrix M=get<matrix_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<matrix_value>
      (lattice::eigen_lattice(M->val,eigen_value)));
}

void row_saturate_wrapper(expression_base::level l)
{ shared_matrix M=get<matrix_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<matrix_value>(lattice::row_saturate(M->val)));
}

void invfact_wrapper(expression_base::level l)
{ shared_matrix m=get<matrix_value>();
  if (l==expression_base::no_value)
    return;
  own_vector inv_factors = std::make_shared<vector_value>(std::vector<int>());
matreduc::Smith_basis(m->val,inv_factors->val);
  push_value(std::move(inv_factors));
}

void Smith_basis_wrapper(expression_base::level l)
{ shared_matrix m=get<matrix_value>();
  if (l==expression_base::no_value)
    return;
  own_vector inv_factors = std::make_shared<vector_value>(std::vector<int>());
push_value(std::make_shared<matrix_value>
    (matreduc::Smith_basis(m->val,inv_factors->val)));
}

void Smith_wrapper(expression_base::level l)
{ shared_matrix m=get<matrix_value>();
  if (l==expression_base::no_value)
    return;
  own_vector inv_factors = std::make_shared<vector_value>(std::vector<int>());
push_value(std::make_shared<matrix_value>
    (matreduc::Smith_basis(m->val,inv_factors->val)));
  push_value(std::move(inv_factors));
  if (l==expression_base::single_value)
    wrap_tuple<2>();
}

void invert_wrapper(expression_base::level l)
{ shared_matrix m=get<matrix_value>();
  if (m->val.numRows()!=m->val.numColumns())
  { std::ostringstream s;
    s<< "Cannot invert a " 
     << m->val.numRows() << "x" << m->val.numColumns() << " matrix";
    throw runtime_error(s.str());
  }
  if (l==expression_base::no_value)
    return;
  own_int denom = std::make_shared<int_value>(0);
push_value(std::make_shared<matrix_value>(m->val.inverse(denom->val)));
  push_value(std::move(denom));
  if (l==expression_base::single_value)
    wrap_tuple<2>();
}

void section_wrapper(expression_base::level l)
{
  shared_matrix m=get<matrix_value>();
  BinaryMap A(m->val);
  BinaryMap B=A.section();
  own_matrix res = std::make_shared<matrix_value>(
    int_Matrix(B.numRows(),B.numColumns()));
  for (unsigned int j=B.numColumns(); j-->0;)
    res->val.set_column(j,int_Vector(B.column(j)));
  if (l!=expression_base::no_value)
    push_value(std::move(res));
}

void subspace_normal_wrapper(expression_base::level l)
{
  typedef BitVector<64> bitvec;
  shared_matrix generators=get<matrix_value>();
  unsigned int n_gens = generators->val.numColumns();
  unsigned int dim = generators->val.numRows();
  if (dim>64)
    throw runtime_error("Dimension too large: "+str(dim)+">64");
  if (n_gens>64)
    throw runtime_error ("Too many generators: "+str(n_gens)+">64");

  std::vector<bitvec> basis, combination;
    // |basis[j]| will be initialised from column $j$ of |generators|
  std::vector<unsigned int> pivot; // |pivot[i]| is bit position for |basis[i]|
  std::vector<unsigned int> pivoter;
    // generator |pivoter[i]| led to |basis[i]|
  { unsigned max_rank=std::min(n_gens,dim); // |basis| cannot exceed this size
    basis.reserve(max_rank); pivot.reserve(max_rank); pivoter.reserve(max_rank);
  }
  bitvector::initBasis(combination,n_gens);
    // express (still virtual) |basis| elements in |generators|

  
  for (unsigned int j=0; j<n_gens; ++j)
  { bitvec v(generators->val.column(j)); // reduce modulo $2$
    for (unsigned int l=0; l<basis.size(); ++l)
      if (v[pivot[l]])
      {
         v -= basis[l];
         combination[j] -= combination[pivoter[l]];
      }
    if (v.nonZero())
    {
      unsigned int piv = v.firstBit(); // new pivot
      for (unsigned int l=0; l<basis.size(); ++l)
        if (basis[l][piv])
        {
           basis[l] -= v;
           combination[pivoter[l]] -= combination[j];
        }
      basis.push_back(v);
      pivoter.push_back(j);
      pivot.push_back(piv);
    }
  }
  if (l==expression_base::no_value)
    return;

  
  { Permutation pi = permutations::standardization(pivot,dim);
      // relative positions of pivots
    unsigned rank=basis.size(); // dimension of the subspace
    int_Matrix basis_m(dim,rank,0);
    int_Matrix combin_m(n_gens,rank,0);
    int_Matrix relations_m(n_gens,n_gens-rank);
    own_row pivot_r = std::make_shared<row_value>(rank);
    unsigned int l=0; // number of basis vectors copied so far, current index
    for (unsigned int j=0; j<n_gens; ++j)
      if (l<rank and j==pivoter[l])
      { unsigned d = pi[l]; // destination position
        for (auto it=basis[l].data().begin(); it(); ++it)
          basis_m(*it,d) = 1;
        for (auto it= combination[j].data().begin(); it(); ++it)
          combin_m(*it,d) = 1;
        pivot_r->val[d] = std::make_shared<int_value>(pivot[l]);
        ++l;
      }
      else
      { unsigned d = j-l;
        for (auto it= combination[j].data().begin(); it(); ++it)
          relations_m(*it,d) = 1;
      }
    assert (l==rank);
  push_value(std::make_shared<matrix_value>(std::move(basis_m)));
    push_value(std::make_shared<matrix_value>(std::move(combin_m)));
    push_value(std::make_shared<matrix_value>(std::move(relations_m)));
    push_value(std::move(pivot_r));
  }
  if (l==expression_base::single_value)
    wrap_tuple<4>();
}

}

void Id_table::add(id_type id, shared_value val, type_expr&& type, bool is_const)
{ auto its = table.equal_range(id);

  if (its.first==its.second) // no global identifier was previously known
  {
#ifdef incompletecpp11
    auto it = table.insert(its.first,std::make_pair(id,id_data()));
      // create a slot
    it->second = id_data 
          (std::make_shared<shared_value>(std::move(val))
          , std::move(type), is_const );
#else
    table.emplace_hint(its.first,id, id_data 
    (std::make_shared<shared_value>(std::move(val)),std::move(type),is_const));
#endif
  }
  else // a global identifier was previously known
    its.first->second = id_data(
      std::make_shared<shared_value>(std::move(val)), std::move(type),is_const);
}

void Id_table::add_type_def(id_type id, type_expr&& type)
{ auto its = table.equal_range(id);

  if (its.first==its.second) // no global identifier was previously known
  {
#ifdef incompletecpp11
    auto it = table.insert(its.first,std::make_pair(id,id_data()));
      // create a slot
    it->second = id_data (shared_share(),std::move(type),true);
      // and fill it with |type| only
#else
    table.emplace_hint 
      (its.first,id,id_data(shared_share(),std::move(type),true));
#endif
  }
  else // a global identifier was previously known, replace it
    its.first->second = id_data(shared_share(),std::move(type),true);
}

bool Id_table::is_defined_type(id_type id) const
{ map_type::const_iterator p = table.find(id);
 return p!=table.end() and p->second.value()==nullptr;
}

bool Id_table::remove(id_type id)
{ map_type::iterator p = table.find(id);
  if (p==table.end())
    return false;
  table.erase(p); return true;
}

const_type_p Id_table::type_of(id_type id,bool& is_const) const
{ map_type::const_iterator p=table.find(id);
  if (p==table.end())
    return nullptr;
  is_const=p->second.is_const();
  return &p->second.type();
}
void Id_table::specialise(id_type id,const type_expr& type)
{ map_type::iterator p=table.find(id);
  p->second.type().specialise(type);
}

shared_value Id_table::value_of(id_type id) const
{ map_type::const_iterator p=table.find(id);
  return p==table.end() ? shared_value(value(nullptr)) : *p->second.value();
}
shared_share Id_table::address_of(id_type id)
{ map_type::iterator p=table.find(id);
  if (p==table.end())
    throw logic_error 
    (std::string("Identifier without table entry:")
     +main_hash_table->name_of(id));

  return p->second.value();
}

void Id_table::print(std::ostream& out) const
{ for (map_type::const_iterator p=table.begin(); p!=table.end(); ++p)
  { out << main_hash_table->name_of(p->first);
    const shared_share& v= p->second.value();
    if (v==nullptr)
      out << " = " << p->second.type();
    else
    { out << ": " << p->second.type() << ": ";
      if (*v==nullptr)
        out << '*';
      else
        out << **p->second.value();
     }
     out << std::endl;
   }
}

std::ostream& operator<< (std::ostream& out, const Id_table& p)
{ p.print(out); return out; }

const overload_table::variant_list& overload_table::variants
  (id_type id) const
{ static const variant_list empty;
  auto p=table.find(id);
  return p==table.end() ? empty : p->second;
}

void overload_table::add
  (id_type id, shared_function val, type_expr&& t)
{ assert (t.kind==function_type);
  func_type type(std::move(*t.func)); // steal the function type
  auto its = table.equal_range(id);
  if (its.first==its.second) // a fresh overloaded identifier
  {
#ifdef incompletecpp11
    auto it=table.insert
      (its.first,std::make_pair(id, variant_list()));
#else
    auto it=table.emplace_hint(its.first,id,variant_list())
;
#endif
    it->second.push_back(
      overload_data( std::move(val), std::move(type)) );
  }
  else
  { variant_list& slot=its.first->second; // vector of all variants
    
    { size_t lwb=0; size_t upb=slot.size();
      for (size_t i=0; i<slot.size(); ++i)
      { unsigned int cmp= is_close(type.arg_type,slot[i].type().arg_type);
        switch (cmp)
        {
          case 0x6: lwb=i+1; break;
            // existent type |i| converts to |type|, which must come later
          case 0x5:  if (upb>i) upb=i;  break;
            // |type| converts to type |i|, so it must come before
          case 0x7: // mutually convertible types, maybe identical ones
            if (slot[i].type().arg_type==type.arg_type)
              // identical ones: overload redefinition case
            { slot[i] = overload_data(std::move(val),std::move(type));
                return;
              }
          // |else| {\bf fall through}
          case 0x4:
             
             { std::ostringstream o;
               o << "Cannot overload `" << main_hash_table->name_of(id) << "': " 
                    "previous type " << slot[i].type().arg_type
                 << "\nis too close to " << type.arg_type
                 << ",\nmaking overloading potentially ambiguous." 
                    " Broadness cannot disambiguate,\nas "
                 << (cmp==0x4 ? "neither" :"either") 
                 << " type converts to the other";
               throw program_error(o.str());
             }
          default: {} // nothing for unrelated argument types
        }
      }
      
      if (lwb>upb)
        throw logic_error("Conflicting order of related overload types");
      else
      { slot.insert // better use |emplace| with gcc 4.8
        (slot.begin()+upb,overload_data(std::move(val),std::move(type)));
      }
    }
  }
}

bool overload_table::remove(id_type id, const type_expr& arg_t)
{ map_type::iterator p=table.find(id);
  if (p==table.end()) return false; // |id| was not known at all
  variant_list& variants=p->second;
  for (size_t i=0; i<variants.size(); ++i)
    if (variants[i].type().arg_type==arg_t)
    {
      variants.erase(variants.begin()+i);
      return true;
    }
  return false; // |id| was known, but no such overload is present
}

void overload_table::print(std::ostream& out) const
{ for (auto p=table.begin(); p!=table.end(); ++p)
    for (auto it=p->second.begin(); it!=p->second.end(); ++it)
      out << main_hash_table->name_of(p->first) << ": " 
        << it->type() << ": " << *it->value() << std::endl;
}

std::ostream& operator<< (std::ostream& out, const overload_table& p)
{ p.print(out); return out; }

void initialise_evaluator()
{ 
  execution_stack.reserve(16); 
  
  coercion(int_type,rat_type, "QI", rational_convert); 
  coercion(row_of_rat_type,ratvec_type, "Qv[Q]", ratlist_ratvec_convert); 
  coercion(ratvec_type,row_of_rat_type, "[Q]Qv", ratvec_ratlist_convert); 
  coercion(vec_type,ratvec_type,"QvV", vec_ratvec_convert); 
  coercion(row_of_int_type,ratvec_type,"Qv[I]", intlist_ratvec_convert);
  
  coercion(row_of_int_type, vec_type, "V[I]", intlist_vector_convert); 
  coercion(row_of_vec_type,mat_type, "M[V]", veclist_matrix_convert); 
  coercion(row_row_of_int_type,mat_type, "M[[I]]", intlistlist_matrix_convert); 
  coercion(vec_type,row_of_int_type, "[I]V", vector_intlist_convert); 
  coercion(mat_type,row_of_vec_type, "[V]M", matrix_veclist_convert); 
  coercion(mat_type,row_row_of_int_type, "[[I]]M", matrix_intlistlist_convert); 
  
  install_function(plus_wrapper,"+","(int,int->int)");
  install_function(minus_wrapper,"-","(int,int->int)");
  install_function(times_wrapper,"*","(int,int->int)");
  install_function(divide_wrapper,"\\","(int,int->int)");
  install_function(modulo_wrapper,"%","(int,int->int)");
  install_function(divmod_wrapper,"\\%","(int,int->int,int)");
  install_function(unary_minus_wrapper,"-","(int->int)");
  install_function(power_wrapper,"^","(int,int->int)");
  install_function(fraction_wrapper,"/","(int,int->rat)");
  install_function(unfraction_wrapper,"%","(rat->int,int)");
     // unary \% means ``break open''
  install_function(rat_plus_int_wrapper,"+","(rat,int->rat)");
  install_function(rat_minus_int_wrapper,"-","(rat,int->rat)");
  install_function(rat_times_int_wrapper,"*","(rat,int->rat)");
  install_function(rat_divide_int_wrapper,"/","(rat,int->rat)");
  install_function(rat_modulo_int_wrapper,"%","(rat,int->rat)");
  install_function(rat_plus_wrapper,"+","(rat,rat->rat)");
  install_function(rat_minus_wrapper,"-","(rat,rat->rat)");
  install_function(rat_times_wrapper,"*","(rat,rat->rat)");
  install_function(rat_divide_wrapper,"/","(rat,rat->rat)");
  install_function(rat_modulo_wrapper,"%","(rat,rat->rat)");
  install_function(rat_unary_minus_wrapper,"-","(rat->rat)");
  install_function(rat_inverse_wrapper,"/","(rat->rat)");
  install_function(rat_power_wrapper,"^","(rat,int->rat)");
  install_function(int_unary_eq_wrapper,"=","(int->bool)");
  install_function(int_unary_neq_wrapper,"!=","(int->bool)");
  install_function(int_non_negative_wrapper,">=","(int->bool)");
  install_function(int_positive_wrapper,">","(int->bool)");
  install_function(int_non_positive_wrapper,"<=","(int->bool)");
  install_function(int_negative_wrapper,"<","(int->bool)");
  install_function(int_eq_wrapper,"=","(int,int->bool)");
  install_function(int_neq_wrapper,"!=","(int,int->bool)");
  install_function(int_less_wrapper,"<","(int,int->bool)");
  install_function(int_lesseq_wrapper,"<=","(int,int->bool)");
  install_function(int_greater_wrapper,">","(int,int->bool)");
  install_function(int_greatereq_wrapper,">=","(int,int->bool)");
  install_function(rat_unary_eq_wrapper,"=","(rat->bool)");
  install_function(rat_unary_neq_wrapper,"!=","(rat->bool)");
  install_function(rat_non_negative_wrapper,">=","(rat->bool)");
  install_function(rat_positive_wrapper,">","(rat->bool)");
  install_function(rat_non_positive_wrapper,"<=","(rat->bool)");
  install_function(rat_negative_wrapper,"<","(rat->bool)");
  install_function(rat_eq_wrapper,"=","(rat,rat->bool)");
  install_function(rat_neq_wrapper,"!=","(rat,rat->bool)");
  install_function(rat_less_wrapper,"<","(rat,rat->bool)");
  install_function(rat_lesseq_wrapper,"<=","(rat,rat->bool)");
  install_function(rat_greater_wrapper,">","(rat,rat->bool)");
  install_function(rat_greatereq_wrapper,">=","(rat,rat->bool)");
  install_function(equiv_wrapper,"=","(bool,bool->bool)");
  install_function(inequiv_wrapper,"!=","(bool,bool->bool)");
  install_function(string_unary_eq_wrapper,"=","(string->bool)");
  install_function(string_unary_neq_wrapper,"!=","(string->bool)");
  install_function(string_eq_wrapper,"=","(string,string->bool)");
  install_function(string_neq_wrapper,"!=","(string,string->bool)");
  install_function(string_less_wrapper,"<","(string,string->bool)");
  install_function(string_leq_wrapper,"<=","(string,string->bool)");
  install_function(string_greater_wrapper,">","(string,string->bool)");
  install_function(string_geq_wrapper,">=","(string,string->bool)");
  install_function(string_concatenate_wrapper,"##","(string,string->string)");
  install_function(concatenate_strings_wrapper,"##","([string]->string)");
  install_function(int_format_wrapper,"int_format","(int->string)");
  install_function(string_to_ascii_wrapper,"ascii","(string->int)");
  install_function(ascii_char_wrapper,"ascii","(int->string)");
  install_function(sizeof_string_wrapper,"#","(string->int)");
  install_function(sizeof_vector_wrapper,"#","(vec->int)");
  install_function(sizeof_ratvec_wrapper,"#","(ratvec->int)");
  install_function(matrix_bounds_wrapper,"#","(mat->int,int)");
  install_function(vector_suffix_wrapper,"#","(vec,int->vec)");
  install_function(vector_prefix_wrapper,"#","(int,vec->vec)");
  install_function(join_vectors_wrapper,"##","(vec,vec->vec)");
  install_function(join_vector_row_wrapper,"##","([vec]->vec)");
  install_function(vec_unary_eq_wrapper,"=","(vec->bool)");
  install_function(vec_unary_neq_wrapper,"!=","(vec->bool)");
  install_function(vec_eq_wrapper,"=","(vec,vec->bool)");
  install_function(vec_neq_wrapper,"!=","(vec,vec->bool)");
  install_function(vec_non_negative_wrapper,">=","(vec->bool)");
  install_function(vec_positive_wrapper,">","(vec->bool)");
  install_function(ratvec_unary_eq_wrapper,"=","(ratvec->bool)");
  install_function(ratvec_unary_neq_wrapper,"!=","(ratvec->bool)");
  install_function(ratvec_non_negative_wrapper,">=","(ratvec->bool)");
  install_function(ratvec_positive_wrapper,">","(ratvec->bool)");
  install_function(ratvec_eq_wrapper,"=","(ratvec,ratvec->bool)");
  install_function(ratvec_neq_wrapper,"!=","(ratvec,ratvec->bool)");
  install_function(mat_unary_eq_wrapper,"=","(mat->bool)");
  install_function(mat_unary_neq_wrapper,"!=","(mat->bool)");
  install_function(mat_eq_wrapper,"=","(mat,mat->bool)");
  install_function(mat_neq_wrapper,"!=","(mat,mat->bool)");
  install_function(vec_plus_wrapper,"+","(vec,vec->vec)");
  install_function(vec_minus_wrapper,"-","(vec,vec->vec)");
  install_function(vec_unary_minus_wrapper,"-","(vec->vec)");
  install_function(vec_times_int_wrapper,"*","(vec,int->vec)");
  install_function(vec_divide_int_wrapper,"\\","(vec,int->vec)");
  install_function(vec_modulo_int_wrapper,"%","(vec,int->vec)");
  install_function(vector_div_wrapper,"/","(vec,int->ratvec)");
  install_function(ratvec_unfraction_wrapper,"%","(ratvec->vec,int)");
  install_function(ratvec_plus_wrapper,"+","(ratvec,ratvec->ratvec)");
  install_function(ratvec_minus_wrapper,"-","(ratvec,ratvec->ratvec)");
  install_function(ratvec_unary_minus_wrapper,"-","(ratvec->ratvec)");
  install_function(ratvec_times_int_wrapper,"*","(ratvec,int->ratvec)");
  install_function(ratvec_divide_int_wrapper,"/","(ratvec,int->ratvec)");
  install_function(ratvec_modulo_int_wrapper,"%","(ratvec,int->ratvec)");
  install_function(ratvec_times_rat_wrapper,"*","(ratvec,rat->ratvec)");
  install_function(ratvec_divide_rat_wrapper,"/","(ratvec,rat->ratvec)");
  install_function(mat_plus_int_wrapper,"+","(mat,int->mat)");
  install_function(mat_minus_int_wrapper,"-","(mat,int->mat)");
  install_function(int_plus_mat_wrapper,"+","(int,mat->mat)");
  install_function(int_minus_mat_wrapper,"-","(int,mat->mat)");
  install_function(vv_prod_wrapper,"*","(vec,vec->int)");
  install_function(flex_add_wrapper,"flex_add","(vec,vec->vec)");
  install_function(flex_sub_wrapper,"flex_sub","(vec,vec->vec)");
  install_function(vector_convolve_wrapper,"convolve","(vec,vec->vec)");
  install_function(mat_plus_mat_wrapper,"+","(mat,mat->mat)");
  install_function(mat_minus_mat_wrapper,"-","(mat,mat->mat)");
  install_function(mrv_prod_wrapper,"*","(mat,ratvec->ratvec)");
  install_function(mv_prod_wrapper,"*","(mat,vec->vec)");
  install_function(mm_prod_wrapper,"*","(mat,mat->mat)");
  install_function(vm_prod_wrapper,"*","(vec,mat->vec)");
  install_function(rvm_prod_wrapper,"*","(ratvec,mat->ratvec)");
  
  install_function(null_vec_wrapper,"null","(int->vec)");
  install_function(null_mat_wrapper,"null","(int,int->mat)");
  install_function(transpose_vec_wrapper,"^","(vec->mat)");
  install_function(transpose_mat_wrapper,"^","(mat->mat)");
    // install as operator
  install_function(transpose_mat_wrapper,"transpose ","(mat->mat)");
    // use of space in the name makes this copy untouchable
  install_function(id_mat_wrapper,"id_mat","(int->mat)");
  install_function(diagonal_wrapper,"diagonal","(vec->mat)");
  install_function(stack_rows_wrapper,"stack_rows","([vec]->mat)");
  install_function(combine_columns_wrapper,"#","(int,[vec]->mat)");
  install_function(combine_rows_wrapper,"^","(int,[vec]->mat)");
  install_function(swiss_matrix_knife_wrapper,"swiss_matrix_knife"
      ,"(int,mat,int,int,int,int->mat)");
  install_function(swiss_matrix_knife_wrapper,"matrix slicer"
      ,"(int,mat,int,int,int,int->mat)"); // space make an untouchable copy
  
  install_function(echelon_wrapper,"echelon","(mat->mat,[int])");
  install_function(diagonalize_wrapper,"diagonalize","(mat->vec,mat,mat)");
  install_function(adapted_basis_wrapper,"adapted_basis","(mat->mat,vec)");
  install_function(kernel_wrapper,"kernel","(mat->mat)");
  install_function(eigen_lattice_wrapper,"eigen_lattice","(mat,int->mat)");
  install_function(row_saturate_wrapper,"row_saturate","(mat->mat)");
  install_function(invfact_wrapper,"inv_fact","(mat->vec)");
  install_function(Smith_basis_wrapper,"Smith_basis","(mat->mat)");
  install_function(Smith_wrapper,"Smith","(mat->mat,vec)");
  install_function(invert_wrapper,"invert","(mat->mat,int)");
  install_function(section_wrapper,"mod2_section","(mat->mat)");
  install_function(subspace_normal_wrapper,
     "subspace_normal","(mat->mat,mat,mat,[int])");
 }

type_expr analyse_types(const expr& e,expression_ptr& p)
{ try
  { type_expr type; // this starts out as an |undetermined_type|
    p = convert_expr(e,type);
    return type;
  }
  catch (const type_error& err)
  { std::cerr << "Error during analysis of expression " << e.loc << std::endl;
    std::cerr << err.what() << ":\n  Subexpression " << err.offender
            << ' ' << err.offender.loc
            << "\n  has wrong type: found " << err.actual
            << " while " << err.required << " was needed.\n";

  }
  catch (const balance_error& err)
  { std::cerr << "Error in expression "
              << err.offender << ' ' << err.offender.loc << "\n  " 
              << err.what() ;
    for (auto it=err.variants.wcbegin(); not err.variants.at_end((it)); ++it)
      std::cerr << ( it==err.variants.wcbegin() ? ": { " : ", " ) << *it;
    std::cerr<< " }" << std::endl;
  }
  catch (const expr_error& err)
  { std::cerr << "Error in expression "
              << err.offender << ' ' << err.offender.loc << "\n  " 
              << err.what() << std::endl;
  }
  catch (const program_error& err)
  { std::cerr << "Error during analysis of expression " << e.loc << "\n  " 
                << err.what() << std::endl;
  }
  throw program_error("Expression analysis failed");

}

void global_set_identifier(const raw_id_pat &raw_pat, expr_p raw, int overload)
{ do_global_set(id_pat(raw_pat),*expr_ptr(raw),overload); } // ensure clean-up

void global_set_identifiers(const raw_let_list& d)
{ std::pair<id_pat,expr> pat_expr = zip_decls(d);
  do_global_set(std::move(pat_expr.first),pat_expr.second,1);
}

void global_declare_identifier(id_type id, type_p t)
{ type_ptr saf(t); // ensure clean-up
  type_expr& type=*t;
  
  { unsigned int input_level = main_input_buffer->include_depth();
    *output_stream << std::setw(2*input_level) << "";
  }
  *output_stream << "Declaring identifier '" << main_hash_table->name_of(id)
            << "': " << type << std::endl;
  static const shared_value undefined_value; // holds a null pointer
  global_id_table->add(id,undefined_value,std::move(type),false);
}

void global_forget_identifier(id_type id)
{ *output_stream << "Identifier '" << main_hash_table->name_of(id)
          << (global_id_table->remove(id) ? "' forgotten" : "' not known")
            << std::endl;
}

void global_forget_overload(id_type id, type_p t)
{ type_ptr saf(t); // ensure clean-up
  const type_expr& type=*t;
  *output_stream << "Definition of '" << main_hash_table->name_of(id)
            << '@' << type 
            << (global_overload_table->remove(id,type)
               ? "' forgotten"
               : "' not known")
            << std::endl;
}

void type_define_identifier
  (id_type id, type_p t, raw_id_pat ip, const YYLTYPE& loc)
{ type_ptr saf(t); id_pat fields(ip); // ensure clean-up
  type_expr& type=*t;
  if (not fields.sublist.empty()) // do this before we move from |type|
  { assert(type.kind==tuple_type);
    auto tp_it =wtl_const_iterator(type.tupple);
    unsigned count=0;
    for (auto it=fields.sublist.wcbegin(); not it.at_end();
         ++tp_it,++count,++it)
      if (it->kind==0x1) // field selector present
      { type_expr fun_type(type.copy(),tp_it->copy()); // make projector type
        shared_function projector =
          std::make_shared<projector_value>(type,count,it->name,loc);
        add_overload(it->name,std::move(projector),std::move(fun_type));
      }
  }

  bool redefine = global_id_table->is_defined_type(id);
  if (not redefine)
    
    { const bool p = global_id_table->present(id);
      if (p or not global_overload_table->variants(id).empty())
      { *output_stream << "Cannot define '" << main_hash_table->name_of(id) 
                  << "' as a type; it is in use as " 
                  << (p? "global variable" : "function") << std::endl;
        return;
      }
    }
  
  { unsigned int input_level = main_input_buffer->include_depth();
    *output_stream << std::setw(2*input_level) << "";
  }
  *output_stream << "Type name '" << main_hash_table->name_of(id) 
            << (redefine ? "' redefined as " : "' defined as ") << type
            << std::endl;
global_id_table->add_type_def(id,std::move(type));
}

void type_of_expr(expr_p raw)
{ expr_ptr saf(raw); const expr& e=*raw;
  try
  { expression_ptr p;
    *output_stream << "type: " << analyse_types(e,p) << std::endl;
  }
  catch (std::exception& err) { std::cerr<<err.what()<<std::endl; }
}

void show_overloads(id_type id,std::ostream& out)
{ const overload_table::variant_list& variants =
   global_overload_table->variants(id);
   out
   << (variants.empty() ? "No overloads for '" : "Overloaded instances of '")
 << main_hash_table->name_of(id) << '\'' << std::endl;
 for (size_t i=0; i<variants.size(); ++i)
   out << "  "
    << variants[i].type().arg_type << "->" << variants[i].type().result_type 
    << std::endl;
}

void show_ids(std::ostream& out)
{ out << "Overloaded operators and functions:\n"
      << *global_overload_table 
      << "Global variables:\n" << *global_id_table;
}

void vector_value::print(std::ostream& out) const
{ size_t l=val.size(),w=0; std::vector<std::string> tmp(l);
  for (size_t i=0; i<l; ++i)
  { std::ostringstream s; s<<val[i]; tmp[i]=s.str();
    if (tmp[i].length()>w) w=tmp[i].length();
  }
  if (l==0) out << "[ ]";
  else
  { w+=1; out << std::right << '[';
    for (size_t i=0; i<l; ++i)
      out << std::setw(w) << tmp[i] << (i<l-1 ? "," : " ]");
  }
}

void rational_vector_value::print(std::ostream& out) const
{ size_t l=val.size(),w=0; std::vector<std::string> tmp(l);
  for (size_t i=0; i<l; ++i)
  { std::ostringstream s; s<<val.numerator()[i]; tmp[i]=s.str();
    if (tmp[i].length()>w) w=tmp[i].length();
  }
  if (l==0) out << "[ ]";
  else
  { w+=1; out << std::right << '[';
    for (size_t i=0; i<l; ++i)
      out << std::setw(w) << tmp[i] << (i<l-1 ? "," : " ]");
  }
  out << '/' << val.denominator();
}

void matrix_value::print(std::ostream& out) const
{ size_t k=val.numRows(),l=val.numColumns();
  if (k==0 or l==0)
  {  out << "The " << k << 'x' << l << " matrix"; return; }
  std::vector<size_t> w(l,0);
  for (size_t i=0; i<k; ++i)
    for (size_t j=0; j<l; ++j)
    { std::ostringstream s; s<<val(i,j); size_t len=s.str().length();
      if (len>w[j]) w[j]=len;
    }
  out << std::endl << std::right;
  for (size_t i=0; i<k; ++i)
  { out << '|';
    for (size_t j=0; j<l; ++j)
      out << std::setw(w[j]+1) << val(i,j) << (j<l-1 ? ',' : ' ');
    out << '|' << std::endl;
  }
}

void install_function
 (wrapper_function f,const char*name, const char* type_string)
{ type_ptr type = mk_type(type_string);
  std::ostringstream print_name; print_name<<name;
  if (type->kind!=function_type)
    throw logic_error
     ("Built-in with non-function type: "+print_name.str());
  print_name << '@' << type->func->arg_type;
  auto val = std::make_shared<builtin_value<false> >(f,print_name.str());
  global_overload_table->add
    (main_hash_table->match_literal(name),std::move(val),std::move(*type));
}

void sizeof_string_wrapper(expression_base::level l)
{ size_t s=get<string_value>()->val.size();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(s));
}

void sizeof_vector_wrapper(expression_base::level l)
{ size_t s=get<vector_value>()->val.size();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(s));
}

void sizeof_ratvec_wrapper(expression_base::level l)
{ size_t s=get<rational_vector_value>()->val.size();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(s));
}

void matrix_ncols_wrapper(expression_base::level l)
{ shared_matrix m=get<matrix_value>();
  if (l==expression_base::no_value)
    return;
  push_value(std::make_shared<int_value>(m->val.numColumns()));
}

void mm_prod_wrapper(expression_base::level l)
{ shared_matrix rf=get<matrix_value>(); // right factor
  shared_matrix lf=get<matrix_value>(); // left factor
  check_size(lf->val.numColumns(),rf->val.numRows());
  if (l!=expression_base::no_value)
    push_value(std::make_shared<matrix_value>(lf->val*rf->val));
}

 void transpose_mat_wrapper(expression_base::level l)
{ shared_matrix m=get<matrix_value>();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<matrix_value>(m->val.transposed()));
}

void id_mat_wrapper(expression_base::level l)
{ int i=get<int_value>()->val;
  if (l!=expression_base::no_value)
    push_value(std::make_shared<matrix_value>
      (int_Matrix(std::max(i,0)))); // identity
}

} }

