#include  "parsetree.h"
#include  "axis-types.h" \
   
#include  "lexer.h"

namespace atlas
{ namespace interpreter
  {

source_location::source_location(const YYLTYPE& loc)
: start_line(loc.first_line) // ``narrow'' to |unsigned int|
, extent(loc.last_line-loc.first_line) // narrow to |unsigned short|
, first_col(loc.first_column)
, last_col(loc.last_column) // these are narrowed too
, file(main_input_buffer->current_file())
{}

source_location::source_location
  (const source_location& left, const source_location& right)
: start_line(left.start_line)
, extent(right.start_line-left.start_line+right.extent)
, first_col(left.first_col), last_col(right.last_col)
, file(left.file)
{ // correct the above computations when one location was undefined
  if (right.undefined())
    *this = left;
  else if (left.undefined())
    *this = right;
}

void destroy_expr(expr_p p)  { delete p; }

expr_p make_int_denotation (int val, const YYLTYPE& loc)
{ return new expr(val,loc,expr::int_tag()); }

expr_p make_bool_denotation(bool val, const YYLTYPE& loc)
{ return new expr (val,loc,expr::bool_tag()); }

expr_p make_string_denotation(std::string* val_p, const YYLTYPE& loc)
{ expr_p result=new expr(std::move(*val_p),loc);
  delete val_p;
  return result;
}

expr_p make_applied_identifier (id_type id, const YYLTYPE& loc)
 { return new expr(id,loc,expr::identifier_tag()); }

expr_p make_dollar (const YYLTYPE& loc)
{ return new expr(loc,expr::dollar_tag()); }
expr_p make_break (unsigned n,const YYLTYPE& loc)
{ return new expr(loc,expr::break_tag{n}); }
expr_p make_return (expr_p exp,const YYLTYPE& loc)
{ return new expr(exp,loc,expr::return_tag{}); }
expr_p make_die (const YYLTYPE& loc)
{ return new expr(loc,expr::die_tag()); }

expr::expr(expr_list&& nodes, tuple_display_tag)
: kind(tuple_display)
  , sublist(nodes.release())
  , loc()
{
  if (sublist==nullptr)
    return; // cannot extract something from nothing
  const source_location& left(sublist->contents.loc);
  wel_const_iterator it(sublist);
  while (not std::next(it).at_end())
    ++it;
  loc = source_location(left,it->loc);
}

raw_expr_list make_exprlist_node(expr_p e, raw_expr_list raw)
  { expr_ptr saf(e); expr_list l(raw);
    l.push_front(std::move(*e));
    return l.release();
  }

raw_expr_list reverse_expr_list(raw_expr_list raw)
{ expr_list l(raw); l.reverse(); return l.release(); }

expr_p wrap_tuple_display(raw_expr_list l, const YYLTYPE& loc)
{ return new expr(expr_list(l),expr::tuple_display_tag(),loc); }
expr_p wrap_list_display(raw_expr_list l, const YYLTYPE& loc)
{ return new expr(expr_list(l),expr::list_display_tag(),loc); }

void destroy_exprlist(raw_expr_list l)
{ (expr_list(l)); }

expr_p make_application_node(expr_p f, raw_expr_list r_args,
 const YYLTYPE& loc, const YYLTYPE& left, const YYLTYPE& right)
{ expr_ptr ff(f); expr_list args(r_args);
  app a(new application_node { std::move(*ff), expr() });
  if (args.singleton())
    a->arg.set_from(args.front());
  else
    a->arg=expr(std::move(args),expr::tuple_display_tag(),
      source_location(source_location(left),source_location(right)));
  return new expr(std::move(a),loc); // move construct application expression
}

expr_p make_application_node(expr_p f, expr_p arg, const YYLTYPE& loc)
{ expr_ptr ff(f); expr_ptr args(arg);
  app a(new application_node { std::move(*ff), std::move(*args) } );
  return new expr(std::move(a),loc); // move construct application expression
}

expr_p make_binary_call(id_type name, expr_p x, expr_p y,
   const YYLTYPE& loc, const YYLTYPE& op_loc)
{
  expr_ptr xx(x), yy(y); // wrap in smart pointers for exception safety
  const source_location range(xx->loc,yy->loc); // extent of operands
  expr_list args; // start with empty argument list
  args.push_front(std::move(*yy));
  args.push_front(std::move(*xx)); // build up lest back-to-front
  expr arg_pack(std::move(args),expr::tuple_display_tag(),range);
  app a(new application_node 
    { expr(name,op_loc,expr::identifier_tag()), std::move(arg_pack) });
  return new expr(std::move(a),loc); // move construct application expression
}

expr_p make_unary_call(id_type name, expr_p a,
   const YYLTYPE& loc, const YYLTYPE& op_loc)
{
  expr_ptr aa(a); // wrap in smart pointer for exception safety
  return new expr(new  application_node(
    expr(name,op_loc,expr::identifier_tag()),std::move(*aa)),loc);
}

expr_p make_negation(expr_p e, const YYLTYPE& loc)
{
  return new expr(e,loc,expr::negate_tag());
}

raw_form_stack start_formula
   (expr_p e, id_type op, int prio, const YYLTYPE& op_loc)
{ form_stack result;
  result.emplace_back (
   std::move(*expr_ptr(e)), expr(op,op_loc,expr::identifier_tag()), prio );
  return result.release();
}

raw_form_stack start_unary_formula (id_type op, int prio, const YYLTYPE& op_loc)
{ form_stack result;    // leave |left_subtree| empty
  result.emplace_back (expr(), expr(op,op_loc,expr::identifier_tag()), prio );
  return result.release();
}

raw_form_stack extend_formula
  (raw_form_stack pre, expr_p ep,id_type op, int prio, const YYLTYPE& op_loc)
{ expr e(std::move(*expr_ptr (ep))); form_stack s(pre);
  while (not s.empty() and s.front().prio>=prio+prio%2)
    
    { expr& lt = s.front().left_subtree;
      if (lt.kind==no_expr) // apply initial unary operator
      {
        expr& oper = s.front().op_exp;
        const source_location range(oper.loc,e.loc); // extent of operands
        e = expr(new application_node(std::move(oper),std::move(e)),range);
      }
      else
      {
        expr_list args; // start with empty argument list
        const source_location range(lt.loc,e.loc); // extent of operands
        args.push_front(std::move(e));
        args.push_front(std::move(lt));
        expr arg_pack(std::move(args),expr::tuple_display_tag(),range);
        app a(new application_node 
          ( std::move(s.front().op_exp), std::move(arg_pack) ));
        e= expr(std::move(a),range); // move construct application expression
      }
      s.pop_back();
    }
  s.emplace_back(std::move(e),expr(op,op_loc,expr::identifier_tag()),prio);
  return s.release();
}

expr_p end_formula (raw_form_stack pre, expr_p ep, const YYLTYPE& loc)
{ expr e(std::move(*expr_ptr (ep))); form_stack s(pre);
  while (not s.empty())
    
    { expr& lt = s.front().left_subtree;
      if (lt.kind==no_expr) // apply initial unary operator
      {
        expr& oper = s.front().op_exp;
        const source_location range(oper.loc,e.loc); // extent of operands
        e = expr(new application_node(std::move(oper),std::move(e)),range);
      }
      else
      {
        expr_list args; // start with empty argument list
        const source_location range(lt.loc,e.loc); // extent of operands
        args.push_front(std::move(e));
        args.push_front(std::move(lt));
        expr arg_pack(std::move(args),expr::tuple_display_tag(),range);
        app a(new application_node 
          ( std::move(s.front().op_exp), std::move(arg_pack) ));
        e= expr(std::move(a),range); // move construct application expression
      }
      s.pop_back();
    }
  return new expr(std::move(e));
}

void destroy_formula(raw_form_stack s)
{ (form_stack(s)); }

raw_patlist make_pattern_node(raw_patlist prev,raw_id_pat& pattern)
{ patlist l(prev); l.push_front(id_pat(pattern)); return l.release(); }

void destroy_pattern(raw_patlist p)
{ (patlist(p)); }

void destroy_id_pat(const raw_id_pat& p)
{ if ((p.kind & 0x2)!=0)
    destroy_pattern(p.sublist);
}

raw_patlist reverse_patlist(raw_patlist raw)
{ patlist p(raw);
  p.reverse();
  return p.release();
}

raw_let_list make_let_node(raw_id_pat& pattern, expr_p val)
{ let_list result;
  result.push_front ( let_pair { id_pat(pattern), std::move(*expr_ptr(val)) } );
  return result.release();
}

raw_let_list append_let_node(raw_let_list prev, raw_let_list cur)
{ let_list result(prev);
  result.push_front(std::move(let_list(cur).front()));
  return result.release();
 }

std::pair<id_pat,expr> zip_decls(raw_let_list d)
{ let_list decls(d);
  std::pair<id_pat,expr> result;
  id_pat& pattern=result.first; expr& val=result.second;
  if (decls.singleton()) // single declaration
  { pattern=std::move(decls.front().pattern);
    val=std::move(decls.front().val);
  }
  else
  { patlist patl; expr_list expl;
    for (auto it=decls.begin(); not decls.at_end(it); ++it)
      // zip open |decls|, reversing
    { patl.push_front(std::move(it->pattern));
      expl.push_front(std::move(it->val));
    }
    pattern = id_pat(std::move(patl));
    val=expr(std::move(expl),expr::tuple_display_tag());
      // make a tuple expression
  }
  return result;
}
expr_p make_let_expr_node(raw_let_list d, expr_p b, const YYLTYPE& loc)
{
  expr_ptr bb(b);
  expr& body=*bb; // ensure exception safety
  auto pat_expr = zip_decls(d);
return new expr (new   let_expr_node
  { std::move(pat_expr.first), std::move(pat_expr.second),std::move(body) }
  ,loc);
}

void destroy_letlist(raw_let_list l)
{ (let_list(l)); }

void destroy_type(type_p t) { (type_ptr(t)); }
void destroy_type_list(raw_type_list t) { (type_list(t)); }
  

expr_p make_lambda_node(raw_patlist p, raw_type_list tl, expr_p b,
 const YYLTYPE& loc)
{
  patlist pat_l(p);
  type_list type_l(tl);
  expr_ptr body_p(b);
  expr& body=*body_p; // safety
  id_pat pattern; type_expr parameter_type;
  if (type_l.singleton())
{ pattern=std::move(pat_l.front());
    parameter_type = std::move(type_l.front());
  }
  else
{ pat_l.reverse(); pattern=id_pat(std::move(pat_l));
  type_l.reverse(); parameter_type=type_expr(std::move(type_l));
  // make tuple type
  }
  return new expr(lambda(new lambda_node
      (std::move(pattern),std::move(parameter_type),std::move(body))),loc);
}

expr_p make_conditional_node(expr_p c, expr_p t, expr_p e, const YYLTYPE& loc)
{
  expr_ptr cc(c), tt(t), ee(e);
  expr& cnd=*cc; expr& thn=*tt; expr& els=*ee;
return new  expr(conditional_expr, new 
      conditional_node { std::move(cnd), std::move(thn), std::move(els) },loc);
}

expr_p make_int_case_node(expr_p s, raw_expr_list i, const YYLTYPE& loc)
{
  expr_ptr ss(s); expr_list in_list(i);
  expr& selector=*ss;
  assert(not in_list.empty());
  containers::sl_node<expr>& ins = *i;
  return new expr(int_case_expr,new 
    conditional_node(std::move(selector),std::move(ins)),loc);
}

expr_p make_while_node(expr_p b, unsigned flags, const YYLTYPE& loc)
{
  expr_ptr bb(b); expr& body=*bb;
  return new expr(new while_node { std::move(body), flags},loc) ;
}

expr_p make_for_node
  (raw_id_pat& id, expr_p ip, expr_p b, unsigned flags, const YYLTYPE& loc)
{
  id_pat ind(id); expr_ptr iip(ip), bb(b);
  expr& in=*iip;  expr& body=*bb;
  return new expr(new 
    for_node { std::move(ind), std::move(in), std::move(body), flags },loc);
}

expr_p make_cfor_node
  (id_type id, expr_p c, expr_p l, expr_p b, unsigned flags,
   const YYLTYPE& loc)
{
  expr_ptr cc(c), ll(l), bb(b);
  expr& cnt=*cc; expr& lim=*ll; expr& body=*bb;
  return new expr (new 
    cfor_node { id, std::move(cnt),std::move(lim),std::move(body),flags },loc);
}

expr_p make_subscription_node
  (expr_p a, expr_p i, bool reversed, const YYLTYPE& loc)
{ expr_ptr aa(a); expr_ptr ii(i);
  expr& arr=*aa; expr& ind=*ii;
  return new expr(new
     subscription_node {std::move(arr), std::move(ind), reversed },loc);
}
expr_p make_slice_node
  (expr_p a, expr_p lower, expr_p upper, unsigned flags, const YYLTYPE& loc)
{ expr_ptr aa(a); expr_ptr ll(lower); expr_ptr uu(upper);
expr& arr=*aa; expr& lo=*ll; expr& up=*uu;
  return new expr(new 
     slice_node {std::move(arr), std::move(lo), std::move(up), flags },loc);
}

expr_p make_cast(type_p t, expr_p e, const YYLTYPE& loc)
{
  type_ptr tt(t); expr_ptr ee(e);
  type_expr& type=*tt; expr& exp=*ee;
  return new expr(new cast_node { std::move(type), std::move(exp) },loc);
}

expr_p make_op_cast(id_type name,type_p t, const YYLTYPE& loc)
{
  type_ptr tt(t); type_expr& type=*tt;
  return new expr(new op_cast_node { name, std::move(type) },loc);
}

expr_p make_assignment(id_type id, expr_p r, const YYLTYPE& loc)
{
  expr_ptr rr(r); expr& rhs=*rr;
  return new expr(new  assignment_node { id, std::move(rhs) },loc);
}

expr_p make_multi_assignment(raw_id_pat& lhs, expr_p r, const YYLTYPE& loc)
{
  expr_ptr rr(r); expr& rhs=*rr;
  return new expr(new assignment_node { id_pat(lhs), std::move(rhs) },loc);
}

expr_p make_comp_ass(expr_p l, expr_p r, const YYLTYPE& loc)
{
  expr_ptr ll(l), rr(r);
  assert(ll->kind==subscription); // grammar ensures this
  subscription_node& s = *ll->subscription_variant;
  assert(s.array.kind==applied_identifier); // grammar ensures this
  return new expr(new comp_assignment_node 
  { s.array.identifier_variant
, std::move(s.index)
  , std::move(*rr)
  , s.reversed
  },loc);
}

expr_p make_field_ass(id_type tup, id_type sel, expr_p r, const YYLTYPE& loc)
{
  expr_ptr rr(r);
  return new expr(new field_assignment_node { tup , sel, std::move(*rr) },loc);
}

expr_p make_comp_upd_ass(expr_p l, id_type op, expr_p r,
   const YYLTYPE& loc, const YYLTYPE& op_loc)
{
  expr_ptr ll(l), rr(r);
  assert(ll->kind==subscription); // grammar ensures this
  auto& s = *ll->subscription_variant;
  assert(s.array.kind==applied_identifier); // likewise
  id_type array_name=s.array.identifier_variant;
      // save before move from |s.array|
  id_type hidden=lookup_identifier("$");
  id_pat v(hidden);
  expr::identifier_tag id_tag;
  expr_ptr subs(new expr (new subscription_node
    (std::move(s.array)
    ,expr(hidden,loc,id_tag)
    ,s.reversed
    ),ll->loc));
  expr_ptr formula(make_binary_call(op,subs.release(),rr.release(),loc,op_loc));
  expr ca (new comp_assignment_node 
     {array_name
     ,expr(hidden,loc,id_tag)
     ,std::move(*formula)
     ,s.reversed},loc);
  return new expr(new let_expr_node 
    (std::move(v),std::move(s.index), std::move(ca)),loc);
}

expr_p make_field_upd_ass(expr_p tupex, expr_p selex, id_type op, expr_p r,
   const YYLTYPE& loc,  const YYLTYPE& op_loc)
{
  expr_ptr rr(r);
  expr_ptr tup_exp(tupex);
  expr_ptr sel_exp(selex);
  assert(tup_exp->kind==applied_identifier and
         sel_exp->kind==applied_identifier); // grammar
  id_type tup=tup_exp->identifier_variant, sel=sel_exp->identifier_variant;
   // save, then move
  expr_ptr selec(new expr (new application_node
    (std::move(*sel_exp),std::move(*tup_exp)),loc));
  expr_ptr formula
    (make_binary_call(op,selec.release(),rr.release(),loc,op_loc));
  return new expr(new field_assignment_node(tup,sel,std::move(*formula)),loc);
}

expr_p make_recfun(id_type f, expr_p d,
   const YYLTYPE& loc, const YYLTYPE& f_loc)
{
  expr_ptr dd(d); expr& definition=*dd;
  assert(definition.kind==lambda_expr); // grammar ensures this
  lambda_node& lam = *definition.lambda_variant;
  assert(lam.body.kind==cast_expr); // grammar ensures this
  cast_node& body = *lam.body.cast_variant;
  expr die_expr(f_loc,expr::die_tag());
  expr dummy_body (new cast_node(body.type.copy(),std::move(die_expr)),f_loc);
  expr dummy_f
   (new lambda_node(id_pat(),lam.parameter_type.copy(),std::move(dummy_body))
    ,loc);
  expr rec_assign (new assignment_node(f,std::move(definition)),loc);
  return new expr(new let_expr_node 
    (id_pat(f),std::move(dummy_f),std::move(rec_assign)),loc);
}

expr_p make_sequence
  (expr_p f, expr_p l, unsigned which, const YYLTYPE& loc)
{
  expr_ptr ff(f), ll(l); expr& first=*ff; expr& last=*ll;
  return new expr(new 
    sequence_node { std::move(first), std::move(last) }, which ,loc);
}

id_type lookup_identifier(const char* name)
{ return main_hash_table->match_literal(name); }

void include_file(int skip_seen)
{ if (not main_input_buffer->push_file
          (lex->scanned_file_name(),skip_seen!=0))
    main_input_buffer->close_includes();
     // nested include failure aborts all includes
}


std::ostream& operator<<(std::ostream& out, const source_location& sl)
{
  out << "at " << main_input_buffer->name_of(sl.file) 
      << ':' << sl.start_line << ':' << sl.first_col << '-';
  if (sl.extent>0)
    out << '-' << sl.start_line+sl.extent << ':';
  return out << sl.last_col;
}

expr::~expr()
{
  if (kind!=no_expr)
  {
    switch (kind)
    { 
      case integer_denotation: case boolean_denotation: break;
      case string_denotation:
      #ifdef incompletecpp11
        delete[](str_denotation_variant); break;
      #else
        str_denotation_variant.~basic_string<char>(); break;
      #endif
      
      case applied_identifier:
      case last_value_computed:
      case break_expr:
      case die_expr: break;
      case return_expr: delete return_variant; break;
      
        case tuple_display:
        case list_display: delete sublist; break;
      
      case function_call: delete call_variant; break;
      
      case negation_expr: delete negation_variant; break;
      
      case let_expr: delete let_variant; break;
      
      case lambda_expr: delete lambda_variant; break;
      
      case conditional_expr: delete if_variant; break;
      
      case int_case_expr: delete if_variant; break;
      
      case while_expr: delete while_variant; break;
      case for_expr:   delete for_variant;   break;
      case cfor_expr:  delete cfor_variant;  break;
      
      case subscription: delete subscription_variant; break;
      case slice: delete slice_variant; break;
      
      case cast_expr: delete cast_variant; break;
      
      case op_cast_expr: delete op_cast_variant; break;
      
      case ass_stat: delete assign_variant; break;
      
      case comp_ass_stat: delete comp_assign_variant; break;
      case field_ass_stat: delete field_assign_variant; break;
      
      case seq_expr:
      case next_expr:
      case do_expr:
         delete sequence_variant; break;
       case no_expr: {}
    }
    kind = no_expr;
  }
}

void expr::set_from (expr& other)
{ assert(kind==no_expr);
  switch (kind=other.kind)
    { 
        case integer_denotation:
          int_denotation_variant = other.int_denotation_variant; break;
        case boolean_denotation:
          bool_denotation_variant = other.bool_denotation_variant; break;
        case string_denotation:
      #ifdef incompletecpp11
          str_denotation_variant = other.str_denotation_variant;
          other.str_denotation_variant=nullptr;
      #else
          new (&str_denotation_variant)
          std::string(std::move(other.str_denotation_variant));
          other.str_denotation_variant.~basic_string<char>();
      #endif
        break;
      
      case applied_identifier: identifier_variant=other.identifier_variant; break;
      case break_expr: break_variant=other.break_variant;
      case return_expr: return_variant=other.return_variant;
      case last_value_computed: case die_expr: break;
      
        case tuple_display:
        case list_display: sublist = other.sublist; other.sublist=nullptr; break;
      
         case function_call:
        call_variant=other.call_variant; break;
      
      case negation_expr: negation_variant=other.negation_variant;
      break;
      
       case let_expr:
         let_variant=other.let_variant;
       break;
      
      case lambda_expr: lambda_variant=other.lambda_variant;
      break;
      
      case conditional_expr:
        if_variant=other.if_variant;
      break;
      
      case int_case_expr:
        if_variant=other.if_variant;
      break;
      
        case while_expr: while_variant=other.while_variant; break;
        case for_expr:   for_variant=other.for_variant;     break;
        case cfor_expr:  cfor_variant=other.cfor_variant;   break;
      
      case subscription: subscription_variant=other.subscription_variant; break;
      case slice:        slice_variant       =other.slice_variant;        break;
      
      case cast_expr: cast_variant=other.cast_variant;
      break;
      
      case op_cast_expr: op_cast_variant=other.op_cast_variant; break;
      
      case ass_stat: assign_variant=other.assign_variant; break;
      
      case comp_ass_stat: comp_assign_variant=other.comp_assign_variant; break;
      case field_ass_stat: field_assign_variant=other.field_assign_variant; break;
      
      case seq_expr:
      case next_expr:
      case do_expr:
         sequence_variant=other.sequence_variant; break;
        case no_expr: {}
    }
  other.kind = no_expr;
  loc = other.loc;
}

expr::expr (expr&& other) : kind(no_expr) 
{ set_from(other); }

void expr::operator= (expr&& other)
{
  if (this!=&other)
  { expr tmp(std::move(other)); // move |other| out of destruction's way
    this->~expr(); // now destruct old value
    set_from(tmp);
  }
}

void expr::swap (expr& other)
{ expr tmp(std::move(*this));
  set_from(other);
  other.set_from(tmp);
}

std::ostream& operator<< (std::ostream& out, const expr& e)
{ switch (e.kind)
  { 
    case integer_denotation: out << e.int_denotation_variant; break;
    case boolean_denotation:
       out << std::boolalpha << e.bool_denotation_variant; break;
    case string_denotation:
      out << '"' << e.str_denotation_variant << '"'; break;
    
    case applied_identifier:
      out << main_hash_table->name_of(e.identifier_variant);
    break;
    case last_value_computed: out << '$';  break;
    case break_expr:
    {
      out << " break ";
      if (e.break_variant>0)
        out << e.break_variant << ' ';
    }
      break;
    case return_expr: { out << " return " << *e.return_variant; }
      break;
    case die_expr: out << " die "; break;
    
    case tuple_display:
    {
      const raw_expr_list l=e.sublist;
      wel_const_iterator it(l);
      if (it.at_end())
        out << "()";
      else
      { out << '(' << *it;
        while (not (++it).at_end())
          out << ',' << *it;
        out << ')';
      }
    }
    break;
    case list_display:
    {
      const raw_expr_list l=e.sublist;
      wel_const_iterator it(l);
      if (it.at_end())
        out << "[]";
      else
      { out << '[' << *it;
        while (not (++it).at_end())
          out << ',' << *it;
        out << ']';
      }
    }
    break;
    
    case function_call:
    { const app& a=e.call_variant;
      const expr& fun=a->fun; const expr& arg=a->arg;
      if (fun.kind==applied_identifier) out << fun;
      else out << '(' << fun << ')';
      if (arg.kind==tuple_display) out << arg;
      else out << '(' << arg << ')';
    }
    break;
    
    case negation_expr: out << " not " << *e.negation_variant;
    break;
    
    case let_expr:
    { const let& lexp=e.let_variant;
      out << "let " << lexp->pattern << '=' << lexp->val <<	 " in " << lexp->body;
    }
    break;
    
    case lambda_expr:
    { const lambda& fun=e.lambda_variant;
      if (fun->parameter_type==void_type)
        out << '@';
      else
        out << '(' << fun->parameter_type << ' ' << fun->pattern << ')';
      out << ':' << fun->body;
    }
    break;
    
    case conditional_expr:
    { const cond& c=e.if_variant; out << " if " << c->condition 
      << " then " << c->branches.contents 
      << " else " << c->branches.next->contents << " fi ";
    }
    break;
    
    case int_case_expr:
    { const cond& c=e.if_variant;
      wel_const_iterator it(&c->branches);
      out << " case " << c->condition << " in " << *it;
      for (++it; not it.at_end(); ++it)
        out  << ", " << *it;
      out << " esac ";
    }
    break;
    
    case while_expr:
    { const w_loop& w=e.while_variant;
     out << " while " << w->body << " od ";
    }
    break;
    case for_expr:
    { const f_loop& f=e.for_variant;
      const patlist& pl = f->id.sublist;
    const id_pat& index = pl.front();
      const id_pat& entry = *++pl.begin(); // |pl.size()==2|
    out << " for " << entry;
      if ((index.kind&0x1)!=0)
        out << '@' << index;
      out << " in " << f->in_part
          << (f->flags[0] ? " ~do " : " do ")  << f->body
          << (f->flags[1] ? " ~od " : " od ");
    }
    break;
    case cfor_expr:
    { const cfor_node& c=*e.cfor_variant;
    out << " for ";
      if (c.id!=id_type(-1))
        out << main_hash_table->name_of(c.id);
      out << ": " << c.count;
      if (not is_empty(c.bound))
        out << " from " << c.bound;
      out << (c.flags[0] ? " ~do " : " do ") << c.body
          << (c.flags[1] ? " ~od " : " od ");
    }
    break;
    
    case subscription:
    { const sub& s=e.subscription_variant;
      const expr& i=s->index;
      out << s->array << '[';
      if (i.kind!=tuple_display) out << i;
      else
      {
        wel_const_iterator it (i.sublist);
        out << *it;
        while (not (++it).at_end())
          out << ',' << *it;
     }
      out << ']';
    }
    break;
    case slice:
    { const slc& s=e.slice_variant;
    out << s->array << (s->flags[0] ? "~[" : "[") 
          << s->lower << (s->flags[1] ? "~:" : ":") 
          << s->upper << (s->flags[2] ? "~]" : "]");
    }
    break;
    
    case cast_expr:
    { const cast& c = e.cast_variant;
      out << c->type << ':' << c->exp ;
    }
    break;
    
    case op_cast_expr:
    { const op_cast& c = e.op_cast_variant;
      out << main_hash_table->name_of(c->oper) << '@' << c->type;
    }
    break;
    
    case ass_stat:
    { const assignment& ass = e.assign_variant;
      if ((ass->lhs.kind&0x3)!=0x1)
        out << "set ";
      out << ass->lhs << ":=" << ass->rhs ;
    }
    break;
    
    case comp_ass_stat:
    { const comp_assignment& ass = e.comp_assign_variant;
      out << main_hash_table->name_of(ass->aggr)
          << (ass->reversed ? "~[" : "[") << ass->index <<  "]:="
          << ass->rhs ;
    }
    break;
    case field_ass_stat:
    { const fld_assignment& ass = e.field_assign_variant;
      out << main_hash_table->name_of(ass->aggr)
          << '.' << main_hash_table->name_of(ass->selector) 
          << ":=" << ass->rhs ;
    }
    break;
    
    case seq_expr:
    { const sequence& seq = e.sequence_variant;
      out << seq->first << "; " << seq->last ;
    }
    break;
    case next_expr:
    { const sequence& seq = e.sequence_variant;
      out << seq->first << " next " << seq->last ;
    }
    break;
    case do_expr:
    { const sequence& seq = e.sequence_variant;
      out << seq->first << " do " << seq->last ;
    }
    break;
     case no_expr: {}
  }
  return out;
}

bool is_empty(const expr& e)
  { return e.kind==tuple_display and e.sublist==nullptr; }

std::ostream& operator<< (std::ostream& out, const id_pat& p)
{ if ((p.kind & 0x2)!=0)
  { out << '(';
    if (not p.sublist.empty())
    { auto it = p.sublist.begin();
      out << *it;
      while (not (++it).at_end())
        out << ',' << *it;
    }
    out << ')';
  }
  if ((p.kind&0x3)==0x3) // both parts present
    out << ':';
  if ((p.kind&0x4)!=0) // identifier declared constant
    out << '!';
  if ((p.kind & 0x1)!=0)
    out << main_hash_table->name_of(p.name);
  return out;
}

  }
}

